import unittest
from xml.etree import ElementTree
import pkg_resources
import json

from prov.model import ProvDocument
from prov.tests import examples
from provdbconnector.utils.converter import to_json, from_json, to_provn, from_provn, to_xml, from_xml, form_string
from provdbconnector.exceptions.utils import NoDocumentException, ParseException


class ConverterTests(unittest.TestCase):
    """
    Test the convert class
    """

    def setUp(self):
        # Reading testfiles from prov package according to:
        # http://stackoverflow.com/questions/6028000/python-how-to-read-a-static-file-from-inside-a-package
        #
        #   Assuming your template is located inside your module's package at this path:
        #   <your_package>/templates/temp_file
        #   the correct way to read your template is to use pkg_resources package from setuptools distribution:
        test_resources = {
            'xml': {'package': 'provdbconnector', 'file': '/tests/resources/primer.provx'},
            'json': {'package': 'provdbconnector', 'file': '/tests/resources/primer.json'},
            'provn': {'package': 'provdbconnector', 'file': '/tests/resources/primer.provn'}
        }
        self.test_files = dict(
            (key, pkg_resources.resource_stream(val['package'], val['file'])) for key, val in test_resources.items())
        self.prov_document = examples.primer_example()

    def tearDown(self):
        """
        Close all files
        """
        [self.test_files[k].close() for k in self.test_files.keys()]

    def test_form_string(self):
        """
        Test the convert from string
        """
        result = form_string(self.test_files["json"])
        self.assertIsNotNone(result)
        self.assertIsInstance(result, ProvDocument)

        result = form_string(self.test_files["xml"])
        self.assertIsNotNone(result)
        self.assertIsInstance(result, ProvDocument)

        with self.assertRaises(NotImplementedError):
            form_string(self.test_files["provn"])

        with self.assertRaises(ParseException):
            form_string("A funny string but no xml, json or prov string")

    def test_to_json(self):
        """
        Test the convert to json
        """
        self.assertRaises(NoDocumentException, lambda: to_json())
        json_document = to_json(self.prov_document)
        self.assertIsInstance(json_document, str)
        try:
            json.loads(json_document)
        except ValueError:
            self.fail("Invalid JSON")

    def test_from_json(self):
        """
        Test the convert from json
        """
        self.assertRaises(NoDocumentException, lambda: from_json())
        prov = from_json(self.test_files['json'])
        self.assertIsInstance(prov, ProvDocument)

    def test_to_provn(self):
        """
        Test the convert to prov-n
        """
        self.assertRaises(NoDocumentException, lambda: to_provn())
        provn_document = to_provn(self.prov_document)
        self.assertIsInstance(provn_document, str)
        # Validate that string is in provn format

    def test_from_provn(self):
        """
        Test the convert from prov-n
        """
        self.assertRaises(NoDocumentException, lambda: from_provn())
        # currently the prov lib don't support from_provn

        with self.assertRaises(NotImplementedError):
            from_provn(self.test_files['provn'])

            # self.assertIsInstance(prov, ProvDocument)

    def test_to_xml(self):
        """
        Test the convert to xml
        """
        self.assertRaises(NoDocumentException, lambda: to_xml())
        xml_document = to_xml(self.prov_document)
        self.assertIsInstance(xml_document, str)
        ElementTree.fromstring(xml_document)

    def test_from_xml(self):
        """
        Test the convert from xml
        """
        self.assertRaises(NoDocumentException, lambda: from_xml())
        prov = from_xml(self.test_files['xml'])
        self.assertIsInstance(prov, ProvDocument)



import unittest
from uuid import UUID

import pkg_resources
from prov.model import ProvDocument, ProvAgent, ProvEntity, ProvActivity, QualifiedName, ProvRelation, ProvRecord, ProvBundle

from provdbconnector.tests import examples as examples
from provdbconnector import ProvDb
from provdbconnector.exceptions.database import InvalidOptionsException, NotFoundException
from provdbconnector import Neo4jAdapter, NEO4J_USER, NEO4J_PASS, NEO4J_HOST, NEO4J_BOLT_PORT
from provdbconnector.db_adapters.baseadapter import METADATA_KEY_TYPE_MAP, METADATA_KEY_PROV_TYPE, \
    METADATA_KEY_IDENTIFIER, METADATA_KEY_NAMESPACES, METADATA_KEY_IDENTIFIER_ORIGINAL
from provdbconnector.exceptions.provapi import NoDataBaseAdapterException, InvalidArgumentTypeException


class ProvDbTestTemplate(unittest.TestCase):
    """
    This abstract test class to test the high level function of you database adapter.
    To use this unitest Template extend from this class.

    .. literalinclude:: ../provdbconnector/tests/db_adapters/in_memory/test_simple_in_memory.py
       :linenos:
       :language: python
       :lines: 25-40

    """
    def __init__(self, *args, **kwargs):
        """
        Prevent from execute the test case directly see:

        http://stackoverflow.com/questions/4566910/abstract-test-case-using-python-unittest

        :param args:
        :param kwargs:
        """
        super(ProvDbTestTemplate, self).__init__(*args, **kwargs)
        self.helper = None
        # Kludge alert: We want this class to carry test cases without being run
        # by the unit test framework, so the `run' method is overridden to do
        # nothing.  But in order for sub-classes to be able to do something when
        # run is invoked, the constructor will rebind `run' from TestCase.
        if self.__class__ != ProvDbTestTemplate:
            # Rebind `run' from the parent class.
            self.run = unittest.TestCase.run.__get__(self, self.__class__)
        else:
            self.run = lambda self, *args, **kwargs: None

    def setUp(self):
        """
        Use the setup method to create a provapi instance with you adapter

        .. warning::
            Override this function if you extend this test!
            Otherwise the test will fail.

        :return:
        """
        # this function will never be executed !!!!
        self.provapi = ProvDb()

    def clear_database(self):
        """
        Override this function to clear your database before each test

        :return:
        """
        raise InvalidArgumentTypeException()



    def test_prov_primer_example(self):
        """
        This test try to save and restore a common prov example document

        :return:
        """
        self.clear_database()
        prov_document = examples.primer_example()
        stored_document_id = self.provapi.save_document_from_prov(prov_document)
        stored_document = self.provapi.get_document_as_prov(stored_document_id)

        self.assertEqual(stored_document, prov_document)

    def test_primer_example_alternate(self):
        """
        This test try to save and restore a common prov example document.
        But in a more complex way

        :return:
        """
        self.clear_database()
        prov_document = examples.primer_example_alternate()
        stored_document_id = self.provapi.save_document_from_prov(prov_document)
        stored_document = self.provapi.get_document_as_prov(stored_document_id)

        self.assertEqual(stored_document, prov_document)

    def test_w3c_publication_1(self):
        """
        This test try to save and restore a common prov example document.

        :return:
        """
        self.clear_database()
        prov_document = examples.w3c_publication_1()
        stored_document_id = self.provapi.save_document_from_prov(prov_document)
        stored_document = self.provapi.get_document_as_prov(stored_document_id)

        self.assertEqual(stored_document, prov_document)

    def test_w3c_publication_2(self):
        """
        This test try to save and restore a common prov example document.

        :return:
        """

        self.clear_database()
        prov_document = examples.w3c_publication_2()
        stored_document_id = self.provapi.save_document_from_prov(prov_document)
        stored_document = self.provapi.get_document_as_prov(stored_document_id)

        self.assertEqual(stored_document, prov_document)

    def test_bundles1(self):
        """
        This test try to save and restore a common prov example document.
        With a bundle and some connections inside the bundle.
        This example is also available via `Provstore <https://provenance.ecs.soton.ac.uk/store/documents/114710/>`

        :return:
        """
        self.clear_database()
        prov_document = examples.bundles1()
        stored_document_id = self.provapi.save_document_from_prov(prov_document)
        stored_document = self.provapi.get_document_as_prov(stored_document_id)
        stored_document_unified = stored_document.flattened().unified()
        prov_document_unified= prov_document.flattened().unified()
        self.assertEqual(stored_document_unified, prov_document_unified)

    def test_bundles2(self):
        """
        This test try to save and restore a common prov example document.
        With a bundle and some connections inside the bundle.
        This example is also available via `Provstore <https://provenance.ecs.soton.ac.uk/store/documents/114704/>`

        The main difference to the bundle_1 is that here we have also a mentionOf connection between bundles.
        See PROV-Links spec for more information

        :return:
        """
        self.clear_database()
        prov_document = examples.bundles2()
        stored_document_id = self.provapi.save_document_from_prov(prov_document)
        stored_document = self.provapi.get_document_as_prov(stored_document_id)

        self.assertEqual(stored_document, prov_document)

    def test_collections(self):
        """
        This test try to save and restore a common prov example document.

        :return:
        """

        self.clear_database()
        prov_document = examples.collections()
        stored_document_id = self.provapi.save_document_from_prov(prov_document)
        stored_document = self.provapi.get_document_as_prov(stored_document_id)

        self.assertEqual(stored_document, prov_document)

    def test_long_literals(self):
        """
        This test try to save and restore a common prov example document.

        :return:
        """

        self.clear_database()
        prov_document = examples.long_literals()
        stored_document_id = self.provapi.save_document_from_prov(prov_document)
        stored_document = self.provapi.get_document_as_prov(stored_document_id)

        self.assertEqual(stored_document, prov_document)

    def test_datatypes(self):
        """
        This test try to save and restore a common prov example document.

        :return:
        """

        self.clear_database()
        prov_document = examples.datatypes()
        stored_document_id = self.provapi.save_document_from_prov(prov_document)
        stored_document = self.provapi.get_document_as_prov(stored_document_id)

        self.assertEqual(stored_document, prov_document)


class ProvDbTests(unittest.TestCase):
    """
    This tests are only for the ProvDb itself. You don't have to extend this test in case you want to write your own
    adapter
    """
    maxDiff = None

    def setUp(self):
        """
        Loads the test xml json and provn data

        """
        self.test_resources = {
            'xml': {'package': 'provdbconnector', 'file': '/tests/resources/primer.provx'},
            'json': {'package': 'provdbconnector', 'file': '/tests/resources/primer.json'},
            'provn': {'package': 'provdbconnector', 'file': '/tests/resources/primer.provn'}
        }
        self.test_prov_files = dict((key, pkg_resources.resource_stream(val['package'], val['file'])) for key, val in
                                    self.test_resources.items())
        self.auth_info = {"user_name": NEO4J_USER,
                          "user_password": NEO4J_PASS,
                          "host": NEO4J_HOST + ":" + NEO4J_BOLT_PORT
                          }
        self.provapi = ProvDb(api_id=1, adapter=Neo4jAdapter, auth_info=self.auth_info)

    def clear_database(self):
        session = self.provapi._adapter._create_session()
        session.run("MATCH (x) DETACH DELETE x")

    def tearDown(self):
        """
        Destroy the prov api and remove all data from neo4j
        :return:
        """
        [self.test_prov_files[k].close() for k in self.test_prov_files.keys()]
        self.clear_database()
        del self.provapi

    # Test create instnace
    def test_provapi_instance(self):
        """
        Try to create a test instnace
        :return:
        """
        self.assertRaises(NoDataBaseAdapterException, lambda: ProvDb())
        self.assertRaises(InvalidOptionsException, lambda: ProvDb(api_id=1, adapter=Neo4jAdapter))

        obj = ProvDb(api_id=1, adapter=Neo4jAdapter, auth_info=self.auth_info)
        self.assertIsInstance(obj, ProvDb)
        self.assertEqual(obj.api_id, 1)

        obj = ProvDb(adapter=Neo4jAdapter, auth_info=self.auth_info)
        self.assertIsInstance(obj.api_id, UUID)

    # Methods that automatically convert to ProvDocument
    def test_save_document_from_json(self):
        """
        Try to create a document from a json buffer
        :return:
        """
        self.clear_database()
        json_buffer = self.test_prov_files["json"]
        self.provapi.save_document_from_json(json_buffer)

    def test_get_document_as_json(self):
        """
        try to get the document as json
        :return:
        """
        self.clear_database()
        example = examples.primer_example()
        document_id = self.provapi.save_document_from_prov(example)

        prov_str = self.provapi.get_document_as_json(document_id)
        self.assertIsNotNone(prov_str)
        self.assertIsInstance(prov_str, str)
        prov_document_reverse = ProvDocument.deserialize(content=prov_str, format="json")
        self.assertEqual(prov_document_reverse, example)

    def test_save_document_from_xml(self):
        """
        Try to create a document from xml
        :return:
        """
        self.clear_database()
        json_buffer = self.test_prov_files["xml"]
        self.provapi.save_document_from_json(json_buffer)

    def test_get_document_as_xml(self):
        """
        try to get the document as xml
        :return:
        """
        self.clear_database()
        example = examples.primer_example()
        document_id = self.provapi.save_document_from_prov(example)

        prov_str = self.provapi.get_document_as_xml(document_id)
        self.assertIsNotNone(prov_str)
        self.assertIsInstance(prov_str, str)

        prov_document_reverse = ProvDocument.deserialize(content=prov_str, format="xml")
        self.assertEqual(prov_document_reverse, example)

    def test_save_document_from_provn(self):
        """
        Try to create a document from provn
        :return:
        """
        self.clear_database()
        json_buffer = self.test_prov_files["provn"]
        with self.assertRaises(NotImplementedError):
            self.provapi.save_document_from_provn(json_buffer)

    def test_get_document_as_provn(self):
        """
        Try to get a document in provn
        :return:
        """
        self.clear_database()
        example = examples.primer_example()
        document_id = self.provapi.save_document_from_prov(example)

        prov_str = self.provapi.get_document_as_provn(document_id)
        self.assertIsNotNone(prov_str)
        self.assertIsInstance(prov_str, str)

        # This check throws NotImplementedError, so skip it

        # prov_document_reverse = ProvDocument.deserialize(content=prov_str,format="provn")
        # self.assertEqual(prov_document_reverse, example)

    # Methods with ProvDocument input / output
    def test_save_document(self):
        """
        Try to create a document from a prov instnace
        :return:
        """
        self.clear_database()
        # test prov document input
        example = examples.primer_example()
        document_id = self.provapi.save_document_from_prov(example)
        self.assertIsNotNone(document_id)
        self.assertIsInstance(document_id, str)

        # test invalid options input
        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi.save_document(1)

    def test_save_document_from_prov(self):
        """
        Try to create a primer example document
        :return:
        """
        self.clear_database()
        example = examples.primer_example()
        document_id = self.provapi.save_document_from_prov(example)
        self.assertIsNotNone(document_id)
        self.assertIsInstance(document_id, str)

    def test_save_document_from_prov_alternate(self):
        """
        Try to create a prov_alternative
        :return:
        """
        self.clear_database()
        example = examples.primer_example_alternate()
        document_id = self.provapi.save_document_from_prov(example)
        self.assertIsNotNone(document_id)
        self.assertIsInstance(document_id, str)

    def test_save_document_from_prov_bundles(self):
        """
        Try to create a document with bundles
        :return:
        """
        self.clear_database()
        example = examples.bundles1()
        document_id = self.provapi.save_document_from_prov(example)
        self.assertIsNotNone(document_id)
        self.assertIsInstance(document_id, str)

    def test_save_document_from_prov_bundles2(self):
        """
        Try to create more bundles
        :return:
        """
        self.clear_database()
        example = examples.bundles2()
        document_id = self.provapi.save_document_from_prov(example)
        self.assertIsNotNone(document_id)
        self.assertIsInstance(document_id, str)

    def test_save_document_from_prov_invalid_arguments(self):
        """
        Try to create a prov with some invalid arguments
        :return:
        """
        self.clear_database()
        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi.save_document_from_prov(None)

    def test_get_document_as_prov(self):
        """
        Try to get the document as ProvDocument instnace

        :return:
        """
        self.clear_database()
        example = examples.bundles2()
        document_id = self.provapi.save_document_from_prov(example)

        prov_document = self.provapi.get_document_as_prov(document_id)
        self.assertIsNotNone(prov_document)
        self.assertIsInstance(prov_document, ProvDocument)

        self.assertEqual(prov_document, example)

    def test_get_document_as_prov_invalid_arguments(self):
        """
        Try to get the prov document with invalid arguments

        :return:
        """
        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi.get_document_as_prov()

    def test_save_bundle_invalid_arguments(self):
        """
        Try to create a bundle with invalid arguments
        :return:
        """
        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi._save_bundle_internal(None)

    def test_save_element_invalid(self):
        """
        Test save_element with invalid args

        """

        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi.save_element("Some cool invalid argument")

    def test_save_record(self):
        """
        Test to save a record (a element or a relation)
        """
        self.clear_database()
        doc = examples.primer_example()
        element = list(doc.get_records(ProvActivity)).pop()
        relation = list(doc.get_records(ProvRelation)).pop()

        #save element
        self.provapi.save_record(element)
        db_element = self.provapi.get_element(element.identifier)

        self.assertEqual(db_element,element)

        #Save relation
        self.provapi.save_record(relation)

        # @todo discuss if it a get_relation method is useful, see  https://github.com/DLR-SC/prov-db-connector/issues/45
        # db_relation = self.provapi.get_relation(relation.identifier)
        # self.assertEqual(db_element, element)

    def test_save_record_invalid(self):
        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi.save_record(list())

        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi.save_record(examples.primer_example())

        class InvalidProvRecordExtend(ProvRecord):
            pass

        with self.assertRaises(InvalidArgumentTypeException):
            doc = examples.primer_example()
            identifier = doc.valid_qualified_name("prov:example")
            self.provapi.save_record(InvalidProvRecordExtend(bundle=doc,identifier=identifier))

    def test_save_element(self):
        """
        Try to save a single record without document_di

        """
        self.clear_database()

        doc = examples.primer_example()
        agent = list(doc.get_records(ProvAgent)).pop()
        entity = list(doc.get_records(ProvEntity)).pop()
        activity = list(doc.get_records(ProvActivity)).pop()

        #Try to save the 3 class types
        self.provapi.save_element(agent)
        self.provapi.save_element(entity)
        self.provapi.save_element(activity)

    def test_get_elements(self):
        """
        Test for the get_elements function
        """
        self.clear_database()
        doc = examples.bundles2()
        self.provapi.save_document(doc)

        activities = self.provapi.get_elements(ProvActivity)
        self.assertIsNotNone(activities)
        self.assertIsInstance(activities,ProvDocument)
        self.assertEqual(len(activities.get_records()),0) # document contains only 4 unknown activities

        entities = self.provapi.get_elements(ProvEntity)
        self.assertIsNotNone(entities)
        self.assertIsInstance(entities, ProvDocument)
        self.assertEqual(len(entities.get_records()), 5)

        agents = self.provapi.get_elements(ProvAgent)
        self.assertIsNotNone(agents)
        self.assertIsInstance(agents, ProvDocument)
        self.assertEqual(len(agents.get_records()), 2)

    def test_get_element_invalid(self):
        """
        Test get element with error

        """

        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi.get_element(None)


        with self.assertRaises(NotFoundException):
            doc = ProvDocument()
            name = doc.valid_qualified_name("prov:Some unused name")
            self.provapi.get_element(name)

    def test_get_element(self):
        """
        Try to save a single record without document_id
        and get the record back from the db

        """
        self.clear_database()
        doc = examples.primer_example()
        agent = list(doc.get_records(ProvAgent)).pop()
        entity = list(doc.get_records(ProvEntity)).pop()
        activity = list(doc.get_records(ProvActivity)).pop()

        #Try to save the 3 class types
        agent_id = self.provapi.save_element(agent)
        entity_id = self.provapi.save_element(entity)
        activity_id = self.provapi.save_element(activity)

        self.assertIsInstance(agent_id,QualifiedName)
        self.assertIsInstance(entity_id,QualifiedName)
        self.assertIsInstance(activity_id,QualifiedName)

        agent_restored = self.provapi.get_element(agent_id)
        entity_restored = self.provapi.get_element(entity_id)
        activity_restored = self.provapi.get_element(activity_id)

        self.assertEqual(agent_restored,agent)
        self.assertEqual(entity_restored,entity)
        self.assertEqual(activity_restored,activity)

    def test_save_bundle(self):
        """
        Test the public method to save bundles
        """
        self.clear_database()
        doc = examples.bundles2()
        bundle = list(doc.bundles).pop()

        self.provapi.save_bundle(bundle)

    def test_save_bundle_invalid(self):
        """
        Test the public method to save bundles with invalid arguments
        """
        doc = examples.primer_example()

        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi.save_bundle(doc)

        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi.save_bundle(None)

    def test_get_bundle(self):
        """
        Test the public method to get bundles
        """
        self.clear_database()
        doc = examples.bundles2()
        bundle = list(doc.bundles).pop()

        self.provapi.save_bundle(bundle)

        db_bundle = self.provapi.get_bundle(bundle.identifier)
        self.assertIsNotNone(db_bundle)
        self.assertIsInstance(db_bundle,ProvBundle)
        self.assertEqual(db_bundle,bundle)

    def test_get_bundle_invalid(self):
        """
        Test with invalid arguemnts
        """

        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi.get_bundle("prov:str") #not allowed
        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi.get_bundle(None)  # not allowed

    def test_save_relation_with_unknown_nodes(self):
        """
        Test to create a relation were the start and end node dose not exist
        This should also work
        """
        self.clear_database()
        doc = examples.primer_example()
        relation = list(doc.get_records(ProvRelation)).pop()

        self.provapi.save_relation(relation)

    def test_save_relation_invalid(self):

        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi.save_relation(None)


        with self.assertRaises(InvalidArgumentTypeException):
            doc = examples.primer_example()
            element = list(doc.get_records(ProvEntity)).pop()
            self.provapi.save_relation(element)

    def test_get_metadata_and_attributes_for_record_invalid_arguments(self):
        """
        Try to get attributes and metadata with invalid arguments
        :return:
        """
        with self.assertRaises(InvalidArgumentTypeException):
            self.provapi._get_metadata_and_attributes_for_record(None)

    def test_save_unknown_prov_typ(self):
        """
        Test to prefer non unknown prov type
        """
        self.clear_database()
        doc = examples.prov_db_unknown_prov_typ_example()
        self.provapi.save_document(doc)
        doc_with_entities = self.provapi.get_elements(ProvEntity)
        self.assertEqual(len(doc_with_entities.records), 2)
        self.assertIsInstance(doc_with_entities.records[0], ProvEntity)
        self.assertIsInstance(doc_with_entities.records[1], ProvEntity)

    def test_save_with_override_default_namespace(self):
        """
        Test to support default namespace overrides
        """
        self.clear_database()
        docA = examples.prov_default_namespace_example("docA")
        docB = examples.prov_default_namespace_example("docB")
        self.provapi.save_document(docA)
        self.provapi.save_document(docB)
        doc_with_entities = self.provapi.get_elements(ProvEntity)
        self.assertEqual(len(doc_with_entities.records), 2)

    def test_get_metadata_and_attributes_for_record(self):
        """
        Test the split into metadata / attributes function
        This function separates the attributes and metadata from a prov record
        :return:
        """
        example = examples.prov_api_record_example()

        result = self.provapi._get_metadata_and_attributes_for_record(example.prov_record)
        metadata_result = result.metadata
        attributes_result = result.attributes

        self.assertIsNotNone(result.attributes)
        self.assertIsNotNone(result.metadata)
        self.assertIsInstance(result.attributes, dict)
        self.assertIsInstance(result.metadata, dict)

        self.assertIsNotNone(metadata_result[METADATA_KEY_PROV_TYPE])
        self.assertIsNotNone(metadata_result[METADATA_KEY_IDENTIFIER])
        self.assertIsNotNone(metadata_result[METADATA_KEY_NAMESPACES])
        self.assertIsNotNone(metadata_result[METADATA_KEY_TYPE_MAP])

        # check metadata
        self.assertEqual(example.metadata[METADATA_KEY_PROV_TYPE], metadata_result[METADATA_KEY_PROV_TYPE])
        self.assertEqual(example.metadata[METADATA_KEY_IDENTIFIER], metadata_result[METADATA_KEY_IDENTIFIER_ORIGINAL])
        self.assertEqual(example.metadata[METADATA_KEY_NAMESPACES], metadata_result[METADATA_KEY_NAMESPACES])
        self.assertEqual(example.metadata[METADATA_KEY_TYPE_MAP], metadata_result[METADATA_KEY_TYPE_MAP])

        identifier = example.metadata[METADATA_KEY_IDENTIFIER]
        self.assertEqual(identifier.namespace.uri + identifier.localpart, metadata_result[METADATA_KEY_IDENTIFIER])
        # check attributes
        self.assertEqual(example.expected_attributes, attributes_result)



from functools import reduce
from io import BufferedReader
from provdbconnector.exceptions.utils import ParseException, NoDocumentException

import six
from prov.model import ProvDocument

import logging
log = logging.getLogger(__name__)


def form_string(content):
    """
    Take a string or BufferedReader as argument and transform the string into a ProvDocument

    :param content: Takes a sting or BufferedReader
    :return: ProvDocument
    """
    if isinstance(content, ProvDocument):
        return content
    elif isinstance(content, BufferedReader):
        content = reduce(lambda total, a: total + a, content.readlines())

    if type(content) is six.binary_type:
        content_str = content[0:15].decode()
        if content_str.find("{") > -1:
            return ProvDocument.deserialize(content=content, format='json')
        if content_str.find('<?xml') > -1:
            return ProvDocument.deserialize(content=content, format='xml')
        elif content_str.find('document') > -1:
            return ProvDocument.deserialize(content=content, format='provn')

    raise ParseException("Unsupported input type {}".format(type(content)))


def to_json(document=None):
    """
    Try to convert a ProvDocument into the json representation

    :param document:
    :type document: prov.model.ProvDocument
    :return: Json string of the document
    :rtype: str
    """
    if document is None:
        raise NoDocumentException()
    return document.serialize(format='json')


def from_json(json=None):
    """
    Try to convert a json string into a document

    :param json: The json str
    :type json: str
    :return: Prov Document
    :rtype: prov.model.ProvDocument
    :raise: NoDocumentException
    """
    if json is None:
        raise NoDocumentException()
    return ProvDocument.deserialize(source=json, format='json')


def to_provn(document=None):
    """
    Try to convert a document into a provn representation

    :param document: Prov document to convert
    :type document: prov.model.ProvDocument
    :return: The prov-n str
    :rtype: str
    :raise: NoDocumentException
    """
    if document is None:
        raise NoDocumentException()
    return document.serialize(format='provn')


def from_provn(provn_str=None):
    """
    Try to convert a provn string into a ProvDocument

    :param provn_str: The string to convert
    :type provn_str: str
    :return: The Prov document
    :rtype: ProvDocument
    :raises: NoDocumentException
    """
    if provn_str is None:
        raise NoDocumentException()
    return ProvDocument.deserialize(source=provn_str, format='provn')


def to_xml(document=None):
    """
    Try to convert a document into an xml string

    :param document: The ProvDocument to convert
    :param document: ProvDocument
    :return: The xml string
    :rtype: str
    """
    if document is None:
        raise NoDocumentException()
    return document.serialize(format='xml')


def from_xml(xml_str=None):
    """
    Try to convert a xml string into a ProvDocument

    :param xml_str: The xml string
    :type xml_str: str
    :return: The Prov document
    :rtype: ProvDocument
    """
    if xml_str is None:
        raise NoDocumentException()
    return ProvDocument.deserialize(source=xml_str, format='xml')



import unittest


class ExamplesTest(unittest.TestCase):
    """
    This test is only to load the example and check if the examples are still running

    """

    def test_bundle_example(self):
        """
        Test the bundle example

        """
        import examples.bundle_example

    def test_test_complex_example_with_neo4j(self):
        """
        Test the neo4j example

        """
        import examples.complex_example_with_neo4j

    def test_file_buffer_example(self):
        """
        Test the file buffer example

        """
        import examples.file_buffer_example

    def test_simple_example(self):
        """
        Test the basic example

        """
        import examples.simple_example

    def test_simple_example_influence(self):
        """
        Test the basic example with influence by relation

        """
        import examples.simple_example_influence

    def test_simple_example_with_neo4j(self):
        """
        Test the basic neo4j example

        """
        import examples.simple_example_with_neo4j

    def test_merge_example(self):
        """
        Test the merge example

        """
        import examples.merge_example

    def test_merge_fail_example(self):
        """
        Test the merge fail example

        """
        import examples.merge_fail_example

    def test_horsemeat_example(self):
        """
        Test the merge fail example
        """
        import examples.horsemeat_example


from prov.model import ProvDocument
from provdbconnector import ProvDb
from provdbconnector.db_adapters.in_memory import SimpleInMemoryAdapter

prov_api = ProvDb(adapter=SimpleInMemoryAdapter, auth_info=None)

# create the prov document
prov_document = ProvDocument()
prov_document.add_namespace("ex", "http://example.com")

prov_document.agent("ex:Bob")
prov_document.activity("ex:Alice")

prov_document.association("ex:Alice", "ex:Bob")
# create bundle
b1 = prov_document.bundle("ex:bundle1")
b1.agent("ex:Yoda")

b2 = prov_document.bundle("ex:bundle2")
b2.agent("ex:Jabba the Hutt")

document_id = prov_api.save_document(prov_document)

print(prov_api.get_document_as_provn(document_id))

# Output:
#
# document
#   prefix ex <http://example.com>
#
#   agent(ex:Bob)
#   activity(ex:Alice, -, -)
#   wasAssociatedWith(ex:Alice, ex:Bob, -)
#   bundle ex:bundle2
#     prefix ex <http://example.com>
#
#     agent(ex:Jabba the Hutt)
#   endBundle
#   bundle ex:bundle1
#     prefix ex <http://example.com>
#
#     agent(ex:Yoda)
#   endBundle
# endDocument



class ProvDbException(Exception):
    """
    Base exception class for all api exceptions.
    """
    pass


class NoDataBaseAdapterException(ProvDbException):
    """
    Thrown, if no database adapter argument is passed to the api class.
    """
    pass


class InvalidArgumentTypeException(ProvDbException):
    """
    Thrown, if an invalid argument is passed to any api method.
    """
    pass


class InvalidProvRecordException(ProvDbException):
    """"
    Thrown, if an invalid record is passed to any api method.
    """
    pass



from provdbconnector.tests.db_adapters.test_baseadapter import AdapterTestTemplate
from provdbconnector.tests.test_prov_db import ProvDbTestTemplate
import unittest


def additional_tests():
    from examples.tests.test_examples import ExamplesTest
    return unittest.defaultTestLoader.loadTestsFromTestCase(ExamplesTest)




import os

from neo4j.exceptions import ConfigurationError
from neo4j.graph import Relationship

import provdbconnector.db_adapters.neo4j.cypher_commands as cypher_commands
from provdbconnector.db_adapters.baseadapter import BaseAdapter
from provdbconnector.db_adapters.baseadapter import METADATA_KEY_PROV_TYPE, METADATA_KEY_TYPE_MAP, \
    METADATA_KEY_IDENTIFIER, METADATA_KEY_NAMESPACES

from provdbconnector.exceptions.database import InvalidOptionsException, AuthException, \
    DatabaseException, CreateRecordException, NotFoundException, CreateRelationException, MergeException

from neo4j import GraphDatabase, basic_auth
from prov.constants import PROV_N_MAP
from collections import namedtuple
from provdbconnector.utils.serializer import encode_string_value_to_primitive, encode_dict_values_to_primitive, \
    split_into_formal_and_other_attributes

import logging

logging.getLogger("neo4j.bolt").setLevel(logging.WARN)
log = logging.getLogger(__name__)

NEO4J_USER = os.environ.get('NEO4J_USERNAME', 'neo4j')
NEO4J_PASS = os.environ.get('NEO4J_PASSWORD', 'neo4jneo4j')
NEO4J_HOST = os.environ.get('NEO4J_HOST', 'localhost')
NEO4J_BOLT_PORT = os.environ.get('NEO4J_BOLT_PORT', '7687')
NEO4J_HTTP_PORT = os.environ.get('NEO4J_HTTP_PORT', '7474')

NEO4J_META_PREFIX = "meta:"



class Neo4jAdapter(BaseAdapter):
    """
    This is the neo4j adapter to store prov. data in a neo4j database

    """
    def __init__(self, *args):
        """
        Setup the class

        :param args: None
        """
        super(Neo4jAdapter, self).__init__()
        self.driver = None
        pass

    def _create_session(self):
        """
        Get a session from the driver

        :return: Session
        :rtype Session
        """
        try:
            session = self.driver.session()
        except OSError as e:
            raise AuthException(e)

        return session

    def connect(self, authentication_options):
        """
        The connect method to create a new instance of the db_driver

        :param authentication_options: Username, password, host and encrypted option
        :return: None
        :rtype: None
        :raises: InvalidOptionsException
        """
        if authentication_options is None:
            raise InvalidOptionsException()

        user_name = authentication_options.get("user_name")
        user_pass = authentication_options.get("user_password")
        encrypted = authentication_options.get("encrypted")
        host = authentication_options.get("host")

        if encrypted is None:
            encrypted = False
        if user_name is None or user_pass is None or host is None:
            raise InvalidOptionsException()

        try:
            self.driver = GraphDatabase.driver("bolt://{}".format(host), encrypted=encrypted, auth=basic_auth(user_name, user_pass))

        except ConfigurationError as e:
            raise InvalidOptionsException(e)

        self._create_session()

    @staticmethod
    def _prefix_metadata(metadata):
        """
        Prefix all keys of a dict, only for the neo4j adapter

        :param metadata:
        :return: A dict with prefixed keys
        """
        prefixed_metadata = dict()
        for key, value in metadata.items():
            prefixed_metadata["{meta_prefix}{key}".format(key=key, meta_prefix=NEO4J_META_PREFIX)] = value

        return prefixed_metadata

    @staticmethod
    def _parse_to_primitive_attributes(attributes, prefixed_metadata):
        """
        Convert the dict values and keys into a neo4j friendly type (dict=>json, list,int,float, QualifiedName=>str, datetime=>str)

        :param attributes:
        :param prefixed_metadata:
        :return:
        """
        all_attributes = attributes.copy()
        all_attributes.update(prefixed_metadata)

        db_attributes = dict()
        # transform values
        for key, value in all_attributes.items():
            key_primitive = encode_string_value_to_primitive(key)
            value_primitive = encode_string_value_to_primitive(value)
            db_attributes.update({key_primitive: value_primitive})

        return db_attributes

    @staticmethod
    def _get_attributes_identifiers_cypher_string(key_list):
        """
        This function return a cypher string with all keys as cypher parameters

        :param key_list:
        :return:
        """
        db_attributes_identifiers = map(lambda key: "`{}`: {{`{}`}}".format(key, key), key_list)
        return ",".join(db_attributes_identifiers)

    @staticmethod
    def _get_attributes_set_cypher_string(key_list, cypher_template=cypher_commands.NEO4J_CREATE_NODE_SET_PART):
        """
        Returns a set cypher command for all keys of the keylist

        :param key_list:
        :param cypher_template:
        :return:
        """
        statements = list()
        for key in key_list:
            statements.append(cypher_template.format(attr_name=key))
        return " ".join(statements)

    def save_element(self, attributes, metadata):
        """
        Saves a single record

        :param attributes: The attributes dict
        :type attributes: dict
        :param metadata: The metadata dict
        :type metadata: dict
        :return: The id of the record
        :rtype: str
        """

        metadata = metadata.copy()
        prefixed_metadata = self._prefix_metadata(metadata)

        # setup merge attributes
        (formal_attributes, other_attributes) = split_into_formal_and_other_attributes(attributes, metadata)

        merge_relevant_keys = list()
        merge_relevant_keys.append("meta:{}".format(METADATA_KEY_IDENTIFIER))
        merge_relevant_keys = merge_relevant_keys + list(formal_attributes.keys())

        other_db_attribute_keys = list()
        other_db_attribute_keys = other_db_attribute_keys + list(other_attributes.keys())
        other_db_attribute_keys = other_db_attribute_keys + list(prefixed_metadata.keys())

        # get set statement for non formal attributes
        attr_for_simple_set = other_db_attribute_keys.copy()
        attr_for_simple_set.remove("meta:" + METADATA_KEY_NAMESPACES)
        attr_for_simple_set.remove("meta:" + METADATA_KEY_TYPE_MAP)
        cypher_set_statement = self._get_attributes_set_cypher_string(attr_for_simple_set)

        attr_for_list_merge = list()
        attr_for_list_merge.append("meta:" + METADATA_KEY_NAMESPACES)
        attr_for_list_merge.append("meta:" + METADATA_KEY_TYPE_MAP)
        cypher_set_statement += self._get_attributes_set_cypher_string(attr_for_list_merge,
                                                                       cypher_commands.NEO4J_CREATE_NODE_SET_PART_MERGE_ATTR)

        # get CASE WHEN ... statement to check if a attribute is different
        cypher_merge_check_statement = self._get_attributes_set_cypher_string(attr_for_simple_set,
                                                                              cypher_commands.NEO4J_CREATE_NODE_MERGE_CHECK_PART)

        # get cypher string for the merge relevant attributes
        cypher_merge_relevant_str = self._get_attributes_identifiers_cypher_string(merge_relevant_keys)

        # get prov type
        provtype = metadata[METADATA_KEY_PROV_TYPE]

        # get db_attributes as dict
        db_attributes = self._parse_to_primitive_attributes(attributes, prefixed_metadata)

        session = self._create_session()

        command = cypher_commands.NEO4J_CREATE_NODE_RETURN_ID.format(label=provtype.localpart,
                                                     formal_attributes=cypher_merge_relevant_str,
                                                     set_statement=cypher_set_statement,
                                                     merge_check_statement=cypher_merge_check_statement)
        with session.begin_transaction() as tx:

            result = tx.run(command, dict(db_attributes))

            record_id = None
            merge_success = 0
            for record in result:
                record_id = record["ID"]
                merge_success = record["check"]

            if record_id is None:
                raise CreateRecordException("No ID property returned by database for the command {}".format(command))
            if merge_success == 0:
                tx.success = True
            else:
                tx.success = False
                raise MergeException(
                    "The attributes {other} could not merged into the existing node, All attributes: {all} ".format(
                        other=other_db_attribute_keys, all=db_attributes))

        return str(record_id)

    def save_relation(self, from_node, to_node, attributes, metadata):
        """
        Save a single relation

        :param from_node: The from node as QualifiedName
        :type from_node: QualifiedName
        :param to_node: The to node as QualifiedName
        :type to_node: QualifiedName
        :param attributes: The attributes dict
        :type attributes: dict
        :param metadata: The metadata dict
        :type metadata: dict
        :return: Id of the relation
        :rtype: str
        """

        metadata = metadata.copy()

        prefixed_metadata = self._prefix_metadata(metadata)

        # setup merge attributes
        (formal_attributes, other_attributes) = split_into_formal_and_other_attributes(attributes, metadata)

        merge_relevant_keys = list()
        merge_relevant_keys.append("meta:{}".format(METADATA_KEY_IDENTIFIER))
        merge_relevant_keys = merge_relevant_keys + list(formal_attributes.keys())

        other_db_attribute_keys = list()
        other_db_attribute_keys = other_db_attribute_keys + list(other_attributes.keys())
        other_db_attribute_keys = other_db_attribute_keys + list(prefixed_metadata.keys())

        # get set statement for non formal attributes

        # Remove namespace and type_map from the direct set statement, because this attributes need to be merged
        attr_for_simple_set = other_db_attribute_keys.copy()
        attr_for_simple_set.remove("meta:" + METADATA_KEY_NAMESPACES)
        attr_for_simple_set.remove("meta:" + METADATA_KEY_TYPE_MAP)
        cypher_set_statement = self._get_attributes_set_cypher_string(attr_for_simple_set)

        # Add separate cypher command to merge the namespaces and tpye map into a list
        attr_for_list_merge = list()
        attr_for_list_merge.append("meta:" + METADATA_KEY_NAMESPACES)
        attr_for_list_merge.append("meta:" + METADATA_KEY_TYPE_MAP)
        cypher_set_statement += self._get_attributes_set_cypher_string(attr_for_list_merge,
                                                                       cypher_commands.NEO4J_CREATE_NODE_SET_PART_MERGE_ATTR)

        # get CASE WHEN ... statement to check if a attribute is different
        cypher_merge_check_statement = self._get_attributes_set_cypher_string(attr_for_simple_set,
                                                                              cypher_commands.NEO4J_CREATE_NODE_MERGE_CHECK_PART)

        # get cypher string for the merge relevant attributes
        cypher_merge_relevant_str = self._get_attributes_identifiers_cypher_string(merge_relevant_keys)

        # get db_attributes as dict
        db_attributes = self._parse_to_primitive_attributes(attributes, prefixed_metadata)

        with self._create_session() as session:

            relationtype = PROV_N_MAP[metadata[METADATA_KEY_PROV_TYPE]]

            command = cypher_commands.NEO4J_CREATE_RELATION_RETURN_ID.format(from_identifier=str(from_node),
                                                             to_identifier=str(to_node),
                                                             relation_type=relationtype,
                                                             formal_attributes=cypher_merge_relevant_str,
                                                             merge_check_statement=cypher_merge_check_statement,
                                                             set_statement=cypher_set_statement
                                                             )
            with session.begin_transaction() as tx:

                result = tx.run(command, dict(db_attributes))

                record_id = None
                merge_success = 0
                for record in result:
                    record_id = record["ID"]
                    merge_success = record["check"]

                if record_id is None:
                    raise CreateRelationException("No ID property returned by database for the command {}".format(command))
                if merge_success == 0:
                    tx.success = True
                else:
                    tx.success = False
                    raise MergeException("The attributes {other} could not merged into the existing node ".format(
                        other=other_db_attribute_keys))

            return str(record_id)

    @staticmethod
    def _split_attributes_metadata_from_node(db_node):
        """
        This functions splits a db node back into attributes and metadata, based on the prefix


        :param db_node:
        :type db_node: dict
        :return: namedTuple(attributes,metadata)
        """
        record = namedtuple('Record', 'attributes, metadata')
        # split data
        metadata = {k.replace(NEO4J_META_PREFIX, ""): v for k, v in db_node._properties.items() if
                    k.startswith(NEO4J_META_PREFIX, 0, len(NEO4J_META_PREFIX))}
        attributes = {k: v for k, v in db_node._properties.items() if
                      not k.startswith(NEO4J_META_PREFIX, 0, len(NEO4J_META_PREFIX))}

        # convert a list of namespace into a string if it is only one item
        # @todo Kind of a hack to pass all test, it is also allowed to return a list of JSON encoded strings
        namespaces = metadata[METADATA_KEY_NAMESPACES]
        if isinstance(namespaces, list):
            # If len is 1 return only the raw JSON string
            if len(namespaces) == 1:
                metadata.update({METADATA_KEY_NAMESPACES: namespaces.pop()})

        # convert a list of namespace into a string if it is only one item
        # @todo Kind of a hack to pass all test, it is also allowed to return a list of JSON encoded strings
        # @todo Find out what I hacked here in 2015...
        type_map = metadata[METADATA_KEY_TYPE_MAP]
        if isinstance(type_map, list):
            # If len is 1 return only the raw JSON string
            if len(type_map) == 1:
                metadata.update({METADATA_KEY_TYPE_MAP: type_map.pop()})

        record = record(attributes, metadata)
        return record

    def _get_cypher_filter_params(self, properties_dict, metadata_dict):
        """
        This functions returns a tuple with the cypher_str for the cypher filter and the right parameter names


        :param properties_dict: Search dict
        :param metadata_dict: Seacrh dict
        :return: Tuple(Keys with metadata prefix (if necessary), cypher filter str )
        """
        metadata_dict_prefixed = {"meta:{}".format(k): v for k, v in metadata_dict.items()}

        # Merge the 2 dicts into one
        filter = properties_dict.copy()
        filter.update(metadata_dict_prefixed)

        encoded_params = encode_dict_values_to_primitive(filter)
        cypher_str = self._get_attributes_identifiers_cypher_string(filter.keys())
        return encoded_params, cypher_str

    def get_records_by_filter(self, attributes_dict=None, metadata_dict=None):
        """
        Return the records by a certain filter

        :param attributes_dict: Filter dict
        :type attributes_dict: dict
        :param metadata_dict: Filter dict for metadata
        :type metadata_dict: dict
        :return: list of all nodes and relations that fit the conditions
        :rtype: list(DbRecord and DbRelation)
        """

        if attributes_dict is None:
            attributes_dict = dict()
        if metadata_dict is None:
            metadata_dict = dict()

        (encoded_params, cypher_str) = self._get_cypher_filter_params(attributes_dict, metadata_dict)

        session = self._create_session()
        records = list()
        result_set = session.run(cypher_commands.NEO4J_GET_RECORDS_BY_PROPERTY_DICT.format(filter_dict=cypher_str), encoded_params)
        for result in result_set:
            record = result["re"]

            if record is None:
                raise DatabaseException("Record response should not be None")
            relation_record = self._split_attributes_metadata_from_node(record)
            records.append(relation_record)
        return records

    def get_records_tail(self, attributes_dict=None, metadata_dict=None, depth=None):
        """
        Return all connected nodes form the origin.


        :param attributes_dict: Filter dict
        :type attributes_dict: dict
        :param metadata_dict: Filter dict for metadata
        :type metadata_dict: dict
        :param depth: Max steps
        :return: list of all nodes and relations that fit the conditions
        :rtype: list(DbRecord and DbRelation)
        """

        if attributes_dict is None:
            attributes_dict = dict()
        if metadata_dict is None:
            metadata_dict = dict()

        (encoded_params, cypher_str) = self._get_cypher_filter_params(attributes_dict, metadata_dict)

        depth_str = ""
        if depth is not None:
            depth_str = "1..{max}".format(max=depth)

        session = self._create_session()
        result_set = session.run(cypher_commands.NEO4J_GET_RECORDS_TAIL_BY_FILTER.format(filter_dict=cypher_str, depth=depth_str),
                                 encoded_params)
        records = list()
        for result in result_set:
            record = result["re"]

            if record is None:
                raise DatabaseException("Record response should not be None")
            relation_record = self._split_attributes_metadata_from_node(record)
            records.append(relation_record)

        return records

    def get_bundle_records(self, bundle_identifier):
        """
        Return all records and relations for the bundle


        :param bundle_identifier:
        :return:
        """

        session = self._create_session()
        result_set = session.run(cypher_commands.NEO4J_GET_BUNDLE_RECORDS,
                                 {'meta:{}'.format(METADATA_KEY_IDENTIFIER): str(bundle_identifier)})
        records = list()
        for result in result_set:
            record = result["re"]

            if record is None:
                raise DatabaseException("Record response should not be None")
            relation_record = self._split_attributes_metadata_from_node(record)
            records.append(relation_record)

        return records

    def get_record(self, record_id):
        """
        Try to find the record in the database

        :param record_id:
        :return: DbRecord
        :rtype: DbRecord
        """

        session = self._create_session()
        result_set = session.run(cypher_commands.NEO4J_GET_RECORD_RETURN_NODE, {"record_id": int(record_id)})

        node = None
        for result in result_set:
            if node is not None:
                raise DatabaseException(
                    "get_record should return only one node for the id {}, command {}".format(record_id,
                                                                                              cypher_commands.NEO4J_GET_RECORD_RETURN_NODE))
            node = result["node"]

        if node is None:
            raise NotFoundException("We cant find the node with the id: {}, database command {}".format(record_id,
                                                                                                        cypher_commands.NEO4J_GET_RECORD_RETURN_NODE))

        return self._split_attributes_metadata_from_node(node)

    def get_relation(self, relation_id):
        """
        Get a relation

        :param relation_id:
        :return: The relation
        :rtype: DbRelation
        """

        session = self._create_session()
        result_set = session.run(cypher_commands.NEO4J_GET_RELATION_RETURN_NODE, {"relation_id": int(relation_id)})

        relation = None
        for result in result_set:
            if not isinstance(result["relation"], Relationship):
                raise DatabaseException(
                    " should return only relationship {}, command {}".format(relation_id, cypher_commands.NEO4J_GET_RECORD_RETURN_NODE))

            relation = result["relation"]

        if relation is None:
            raise NotFoundException("We cant find the relation with the id: {}, database command {}".format(relation_id,
                                                                                                            cypher_commands.NEO4J_GET_RECORD_RETURN_NODE))

        return self._split_attributes_metadata_from_node(relation)

    def delete_records_by_filter(self, attributes_dict=None, metadata_dict=None):
        """
        Delete records and relations by a filter


        :param attributes_dict:
        :param metadata_dict:
        :return:
        """

        if attributes_dict is None:
            attributes_dict = dict()
        if metadata_dict is None:
            metadata_dict = dict()

        (encoded_params, cypher_str) = self._get_cypher_filter_params(attributes_dict, metadata_dict)
        session = self._create_session()

        session.run(cypher_commands.NEO4J_DELETE_NODE_BY_PROPERTIES.format(filter_dict=cypher_str), encoded_params)

        return True

    def delete_record(self, record_id):
        """
        Delete a single record


        :param record_id:
        :return:
        """
        session = self._create_session()
        session.run(cypher_commands.NEO4J_DELETE__NODE_BY_ID, {"node_id": int(record_id)})
        return True

    def delete_relation(self, relation_id):
        """
        Delete a single relation


        :param relation_id:
        :return:
        """
        session = self._create_session()
        session.run(cypher_commands.NEO4J_DELETE_RELATION_BY_ID, {"relation_id": int(relation_id)})
        return True



NEO4J_TEST_CONNECTION = """MATCH (n) RETURN count(n) as count"""

# create
NEO4J_CREATE_DOCUMENT_NODE_RETURN_ID = """
CYPHER 3.5
CREATE (node { }) RETURN ID(node) as ID"""
NEO4J_CREATE_NODE_SET_PART = "SET node.`{attr_name}` = {{`{attr_name}`}}"
NEO4J_CREATE_NODE_SET_PART_MERGE_ATTR = "SET node.`{attr_name}` = (CASE WHEN not exists(node.`{attr_name}`) THEN [{{`{attr_name}`}}] ELSE node.`{attr_name}` + {{`{attr_name}`}}  END)"
NEO4J_CREATE_NODE_MERGE_CHECK_PART = """WITH CASE WHEN check = 0 THEN (CASE  WHEN EXISTS(node.`{attr_name}`) AND node.`{attr_name}` <> {{`{attr_name}`}} THEN 1 ELSE 0 END) ELSE 1 END as check , node"""
NEO4J_CREATE_NODE_RETURN_ID = """
                                CYPHER 3.5
                                MERGE (node:{label} {{{formal_attributes}}})
                                WITH 0 as check, node
                                {merge_check_statement}
                                {set_statement}
                                RETURN ID(node) as ID, check """  # args: provType, values
NEO4J_CREATE_RELATION_RETURN_ID = """
                                CYPHER 3.5
                                MATCH
                                    (from{{`meta:identifier`:'{from_identifier}'}}),
                                    (to{{`meta:identifier`:'{to_identifier}'}})
                                MERGE
                                    (from)-[r:{relation_type} {{{formal_attributes}}}]->(to)
                                    WITH 0 as check, r as node
                                    {merge_check_statement}
                                    {set_statement}
                                RETURN
                                    ID(node) as ID, check
                                """  # args: provType, values
# get
NEO4J_GET_RECORDS_BY_PROPERTY_DICT = """
                            CYPHER 3.5 
                            MATCH (d {{{filter_dict}}} )-[r]-(x {{{filter_dict}}})
                            RETURN DISTINCT r as re
                            //Get all nodes that are alone without connections to other nodes
                            UNION
                            MATCH (a {{{filter_dict}}})
                            RETURN DISTINCT a as re
                        """
NEO4J_GET_RECORDS_TAIL_BY_FILTER = """
                            CYPHER 3.5
                            MATCH (x {{{filter_dict}}})-[r *{depth}]-(y)
                            RETURN  DISTINCT y as re
                            UNION
                            MATCH (x {{{filter_dict}}})-[r *{depth}]-(y)
                            WITH REDUCE(output = [], r IN r | output + r) AS flat
                            UNWIND flat as re
                            RETURN DISTINCT re
                        """

NEO4J_GET_BUNDLE_RECORDS = """
                            CYPHER 3.5
                            MATCH (x {`meta:identifier`: {`meta:identifier`}})-[r *1]-(y)
                            WHERE ALL (rel in r WHERE rel.`prov:type` = 'prov:bundleAssociation')
                            RETURN  DISTINCT y as re
                            UNION
                            //get all relations between the nodes
                            MATCH (origin {`meta:identifier`: {`meta:identifier`}})-[r *1]-(x)-[r_return *1]-(y)-[r_2 *1]-(origin {`meta:identifier`: {`meta:identifier`}})
                            WHERE ALL (rel in r WHERE rel.`prov:type` = 'prov:bundleAssociation')
                            AND ALL (rel in r_2 WHERE rel.`prov:type` = 'prov:bundleAssociation')
                            WITH REDUCE(output = [], r IN r_return | output + r) AS flat
                            UNWIND flat as re
                            RETURN DISTINCT re
                            //get all mentionof relations
                            UNION
                            MATCH (bundle_1 {`meta:identifier`: {`meta:identifier`}})-[r *1]-(x)-[r_return *1]-(y)-[r_2 *1]-(bundle_2)
                            WHERE ALL (rel in r WHERE rel.`prov:type` = 'prov:bundleAssociation')
                            AND ALL (rel in r_2 WHERE rel.`prov:type` = 'prov:bundleAssociation')
                            AND ALL (rel in r_return WHERE rel.`meta:prov_type` = 'prov:Mention'  and startNode(rel) = x)
                            WITH REDUCE(output = [], r IN r_return | output + r) AS flat
                            UNWIND flat as re
                            RETURN DISTINCT re

 """

NEO4J_GET_RECORD_RETURN_NODE = """
CYPHER 3.5
MATCH (node) WHERE ID(node)={record_id} RETURN node"""
NEO4J_GET_RELATION_RETURN_NODE = """
CYPHER 3.5
MATCH ()-[relation]-() WHERE ID(relation)={relation_id}  RETURN relation"""

# delete
NEO4J_DELETE__NODE_BY_ID = """
CYPHER 3.5
MATCH (x) Where ID(x) = {node_id} DETACH DELETE x """
NEO4J_DELETE_NODE_BY_PROPERTIES = """
CYPHER 3.5
MATCH (n {{{filter_dict}}}) DETACH DELETE n"""
NEO4J_DELETE_BUNDLE_NODE_BY_ID = """
CYPHER 3.5
MATCH (b) WHERE id(b)=toInt({bundle_id}) DELETE b """
NEO4J_DELETE_RELATION_BY_ID = """
CYPHER 3.5
MATCH ()-[r]-() WHERE id(r) = {relation_id} DELETE r"""



import json
import unittest

from prov.constants import PROV_TYPE,PROV_RECORD_IDS_MAP
from prov.model import ProvDocument
from provdbconnector.db_adapters.baseadapter import BaseAdapter, METADATA_KEY_IDENTIFIER, METADATA_KEY_TYPE_MAP, METADATA_KEY_NAMESPACES, METADATA_KEY_PROV_TYPE
from provdbconnector.exceptions.database import NotFoundException, MergeException
from provdbconnector.tests.examples import base_connector_record_parameter_example, primer_example,\
    base_connector_relation_parameter_example, base_connector_bundle_parameter_example, base_connector_merge_example
from provdbconnector.utils.serializer import encode_dict_values_to_primitive


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    return repr(obj)

def encode_adapter_result_to_excpect(dict_vals):
    """
    This function translate a metadata dict to an expected version of this dict

    :param dict_vals:
    :return:
    """
    copy = dict_vals.copy()
    if type(copy[METADATA_KEY_NAMESPACES]) is list:
        copy.update({METADATA_KEY_NAMESPACES: copy[METADATA_KEY_NAMESPACES].pop()})

    if type(copy[METADATA_KEY_TYPE_MAP]) is list:
        copy.update({METADATA_KEY_TYPE_MAP: copy[METADATA_KEY_TYPE_MAP].pop()})

    return encode_dict_values_to_primitive(copy)

def insert_document_with_bundles(instance, identifier_prefix=""):
    """
    This function creates a full bundle on your database adapter, to prepare the test data

    :param instance: The db_adapter instnace
    :param identifier_prefix: A prefix for the identifiers

    :return: The ids of the records
    """
    args_record = base_connector_record_parameter_example()
    args_bundle = base_connector_bundle_parameter_example()
    doc = ProvDocument()
    doc.add_namespace("ex", "http://example.com")
    # document with 1 record

    args_record["metadata"].update({METADATA_KEY_IDENTIFIER: doc.valid_qualified_name("ex:" + identifier_prefix + str(args_record["metadata"][METADATA_KEY_IDENTIFIER]))})
    doc_record_id = instance.save_element(args_record["attributes"], args_record["metadata"])

    #bundle with 1 record
    args_bundle["metadata"].update({METADATA_KEY_IDENTIFIER: args_bundle["metadata"][METADATA_KEY_IDENTIFIER]})
    bundle_id = instance.save_element(args_bundle["attributes"], args_bundle["metadata"])

    # add relation

    from_record_args = base_connector_record_parameter_example()
    to_record_args = base_connector_record_parameter_example()
    relation_args = base_connector_relation_parameter_example()

    from_label = doc.valid_qualified_name("ex:{}FROM NODE".format(identifier_prefix))
    to_label = doc.valid_qualified_name("ex:{}TO NODE".format(identifier_prefix))
    from_record_args["metadata"][METADATA_KEY_IDENTIFIER] = from_label
    to_record_args["metadata"][METADATA_KEY_IDENTIFIER] = to_label

    from_record_id = instance.save_element(from_record_args["attributes"], from_record_args["metadata"])
    to_record_id = instance.save_element(to_record_args["attributes"], to_record_args["metadata"])


    relation_id = instance.save_relation(from_label, to_label, relation_args["attributes"], relation_args["metadata"])



    return {
        "relation_id": relation_id,
        "from_record_id": from_record_id,
        "to_record_id": to_record_id,
        "bundle_id": bundle_id,
        "doc_record_id": doc_record_id
    }


class AdapterTestTemplate(unittest.TestCase):
    """
    This test class is a template for each database adapter.
    The following example show how you implement the test for your adapter:

        .. literalinclude:: ../provdbconnector/tests/db_adapters/in_memory/test_simple_in_memory.py
           :linenos:
           :language: python
           :lines: 1-25

    """
    maxDiff = None

    def __init__(self, *args, **kwargs):
        """
        Prevent from execute the test case directly see:

        http://stackoverflow.com/questions/4566910/abstract-test-case-using-python-unittest

        :param args:
        :param kwargs:
        """
        super(AdapterTestTemplate, self).__init__(*args, **kwargs)
        self.helper = None
        # Kludge alert: We want this class to carry test cases without being run
        # by the unit test framework, so the `run' method is overridden to do
        # nothing.  But in order for sub-classes to be able to do something when
        # run is invoked, the constructor will rebind `run' from TestCase.
        if self.__class__ != AdapterTestTemplate:
            # Rebind `run' from the parent class.
            self.run = unittest.TestCase.run.__get__(self, self.__class__)
        else:
            self.run = lambda self, *args, **kwargs: None

    def setUp(self):
        """
        Setup the instnace of your database adapter
          .. warning::
            Override this method otherwise all test will fail

        :return:
        """
        # this function will never be executed !!!!
        self.instance = BaseAdapter()

    def clear_database(self):
        """
        This function is to clear your database adapter before each test
        :return:
        """
        pass


    #Create section
    def test_1_save_element(self):
        """
        This test try to save a simple record

        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_1_save_element.svg
           :align: center
           :scale: 50 %

        **Input-Data**

        .. warning::
            This is a json representation of the input data and not the real input data.
            For example metadata.prov_type is a QualifiedName instance
        .. code-block:: json

            {
               "metadata":{
                  "identifier":"<QualifiedName: prov:example_node>",
                  "namespaces":{
                     "ex":"http://example.com",
                     "custom":"http://custom.com"
                  },
                  "prov_type":"<QualifiedName: prov:Activity>",
                  "type_map":{
                     "int value":"int",
                     "date value":"xds:datetime"
                  }
               },
               "attributes":{
                  "ex:individual attribute":"Some value",
                  "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                  "ex:dict value":{
                     "dict":"value"
                  },
                  "ex:list value":[
                     "list",
                     "of",
                     "strings"
                  ],
                  "ex:double value":99.33,
                  "ex:int value":99
               }
            }

        **Output-Data**

        The output is only a id as string

         `4d3cdc76-467d-4db8-89bf-9accc7b27777`



        """
        self.clear_database()
        args = base_connector_record_parameter_example()

        record_id  = self.instance.save_element(args["attributes"], args["metadata"])
        self.assertIsNotNone(record_id)
        self.assertIs(type(record_id), str, "id should be a string ")

    def test_2_save_relation(self):
        """
        This test try to save a simple relation between 2 identifiers


        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_2_save_relation.svg
           :align: center
           :scale: 50 %

        **Input-Data**

        .. warning::
            This is a json representation of the input data and not the real input data.
            For example metadata.prov_type is a QualifiedName instance

        .. code-block:: json

            {

               "from_node":"<QualifiedName: ex:Yoda>",
               "to_node":"<QualifiedName: ex:Luke Skywalker>",
               "metadata":{
                  "prov_type":"<QualifiedName: prov:Mention>",
                  "type_map":{
                     "date value":"xds:datetime",
                     "int value":"int"
                  },
                  "namespaces":{
                     "custom":"http://custom.com",
                     "ex":"http://example.com"
                  },
                  "identifier":"identifier for the relation"
               },
               "attributes":{
                  "ex:list value":[
                     "list",
                     "of",
                     "strings"
                  ],
                  "ex:int value":99,
                  "ex:double value":99.33,
                  "ex:individual attribute":"Some value",
                  "ex:dict value":{
                     "dict":"value"
                  },
                  "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)"
               },
            }

        **Output-Data**

        The output is only the id of the relation as string

         `4d3cdc76-467d-4db8-89bf-9accc7b27778`



        """

        self.clear_database()
        args_relation = base_connector_relation_parameter_example()
        args_records = base_connector_record_parameter_example()


        from_meta = args_records["metadata"].copy()
        from_meta.update({METADATA_KEY_IDENTIFIER: args_relation["from_node"]})
        self.instance.save_element(args_records["attributes"], from_meta)

        to_meta = args_records["metadata"].copy()
        to_meta.update({METADATA_KEY_IDENTIFIER: args_relation["to_node"]})
        self.instance.save_element(args_records["attributes"], to_meta)


        relation_id = self.instance.save_relation(args_relation["from_node"], args_relation["to_node"], args_relation["attributes"], args_relation["metadata"])
        self.assertIsNotNone(relation_id)
        self.assertIs(type(relation_id), str, "id should be a string ")

    @unittest.skip("We should discuss the possibility of relations that are create automatically nodes")
    def test_3_save_relation_with_unknown_records(self):
        """
        This test is to test the creation of a relation where the nodes are not in the database
        :return:
        """
        self.clear_database()
        args_relation = base_connector_relation_parameter_example()


        # Skip the part that creates the from and to node

        relation_id = self.instance.save_relation( args_relation["from_node"],  args_relation["to_node"],
                                                  args_relation["attributes"], args_relation["metadata"])

        self.assertIsNotNone(relation_id)
        self.assertIs(type(relation_id), str, "id should be a string ")

    def test_4_get_record(self):
        """
        Create a record and then try to get it back

        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_4_get_record.svg
           :align: center
           :scale: 50 %

        **Input-Data**

        `"id-333"`

        **Output-Data**

        The output is all connected nodes and there relations

        .. code-block:: json

            [
               {
                  "ex:int value":99,
                  "ex:list value":[
                     "list",
                     "of",
                     "strings"
                  ],
                  "ex:date value":"2005-06-01 13:33:00",
                  "ex:individual attribute":"Some value",
                  "ex:double value":99.33,
                  "ex:dict value":"{\"dict\": \"value\"}"
               },
               {
                  "identifier":"prov:example_node",
                  "prov_type":"prov:Activity",
                  "type_map":"{\"date value\": \"xds:datetime\", \"int value\": \"int\"}",
                  "namespaces":"{\"custom\": \"http://custom.com\", \"ex\": \"http://example.com\"}"
               }
            ]

        """
        self.clear_database()
        args = base_connector_record_parameter_example()

        record_id = self.instance.save_element(args["attributes"], args["metadata"])#

        record_raw = self.instance.get_record(record_id)


        self.assertIsNotNone(record_raw)
        self.assertIsNotNone(record_raw.attributes)
        self.assertIsNotNone(record_raw.metadata)
        self.assertIsInstance(record_raw.metadata, dict)
        self.assertIsInstance(record_raw.attributes, dict)

        attributes_primitive = encode_dict_values_to_primitive(args["attributes"])
        metadata_primitive = encode_dict_values_to_primitive(args["metadata"])

        self.assertEqual(record_raw.attributes, attributes_primitive)
        self.assertEqual(record_raw.metadata, metadata_primitive)

        self.assertIs(type(record_id), str, "id should be a string ")

    def test_5_get_record_not_found(self):
        """
        Try to get a record with a invalid id
        :return:
        """
        self.clear_database()
        with self.assertRaises(NotFoundException):
            self.instance.get_record("99999999")

    def test_6_get_relation(self):
        """
        create a relation between 2 nodes and try to get the relation back


        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_6_get_relation.svg
           :align: center
           :scale: 50 %


        **Input-Data**

        `"id-333"`

        **Output-Data**

        The output is all connected nodes and there relations

        .. code-block:: json

            [
               {
                  "ex:dict value":"{\"dict\": \"value\"}",
                  "ex:int value":99,
                  "ex:individual attribute":"Some value",
                  "ex:date value":"2005-06-01 13:33:00",
                  "ex:list value":[
                     "list",
                     "of",
                     "strings"
                  ],
                  "ex:double value":99.33
               },
               {
                  "identifier":"identifier for the relation",
                  "prov_type":"prov:Mention",
                  "namespaces":"{\"ex\": \"http://example.com\", \"custom\": \"http://custom.com\"}",
                  "type_map":"{\"date value\": \"xds:datetime\", \"int value\": \"int\"}"
               }
            ]

        """
        self.clear_database()
        from_record_args = base_connector_record_parameter_example()
        to_record_args = base_connector_record_parameter_example()
        relation_args = base_connector_relation_parameter_example()

        from_identifier = relation_args["from_node"]
        to_identifier = relation_args["to_node"]
        from_record_args["metadata"][METADATA_KEY_IDENTIFIER] = from_identifier
        to_record_args["metadata"][METADATA_KEY_IDENTIFIER] = to_identifier

        from_record_id = self.instance.save_element(from_record_args["attributes"], from_record_args["metadata"])  #
        to_record_id = self.instance.save_element(to_record_args["attributes"], to_record_args["metadata"])  #

        relation_id = self.instance.save_relation(from_identifier, to_identifier, relation_args["attributes"], relation_args["metadata"])

        relation_raw = self.instance.get_relation(relation_id)

        self.assertIsNotNone(relation_raw)
        self.assertIsNotNone(relation_raw.attributes)
        self.assertIsNotNone(relation_raw.metadata)
        self.assertIsInstance(relation_raw.metadata, dict)
        self.assertIsInstance(relation_raw.attributes, dict)

        attributes_primitive = encode_dict_values_to_primitive(relation_args["attributes"])
        metadata_primitive = encode_dict_values_to_primitive(relation_args["metadata"])

        self.assertEqual(relation_raw.attributes, attributes_primitive)
        self.assertEqual(relation_raw.metadata, metadata_primitive)

    def test_7_get_relation_not_found(self):
        """
        Try to get a not existing relation


        """
        self.clear_database()
        with self.assertRaises(NotFoundException):
            self.instance.get_relation("99999999")

    ##Filter-Section##
    def test_8_get_records_by_filter(self):
        """

        This test is to get a the whole graph without any filter

        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_8_get_records_by_filter.svg
           :align: center
           :scale: 50 %


        **Input-Data**

        We have no input data for the filter function because we want to get the whole graph

        **Output-Data**

        The output is the only node that exist in the database (was also created during the test)

        .. warning::
            This is a json representation of the input data and not the real input data.
            For example metadata.prov_type is a QualifiedName instance
        .. code-block:: json

            {
               "metadata":{
                  "identifier":"<QualifiedName: prov:example_node>",
                  "namespaces":{
                     "ex":"http://example.com",
                     "custom":"http://custom.com"
                  },
                  "prov_type":"<QualifiedName: prov:Activity>",
                  "type_map":{
                     "int value":"int",
                     "date value":"xds:datetime"
                  }
               },
               "attributes":{
                  "ex:individual attribute":"Some value",
                  "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                  "ex:dict value":{
                     "dict":"value"
                  },
                  "ex:list value":[
                     "list",
                     "of",
                     "strings"
                  ],
                  "ex:double value":99.33,
                  "ex:int value":99
               }
            }

        """
        self.clear_database()
        args = base_connector_record_parameter_example()
        base_connector_bundle_parameter_example()

        record_id = self.instance.save_element(args["attributes"], args["metadata"])

        raw_result = self.instance.get_records_by_filter()

        # check restul
        self.assertIsNotNone(raw_result)
        self.assertIsInstance(raw_result, list)
        self.assertEqual(len(raw_result), 1)
        self.assertIsNotNone(raw_result[0].metadata)
        self.assertIsNotNone(raw_result[0].attributes)

        # check if the metadata of the record equals
        attr_dict = encode_dict_values_to_primitive(args["attributes"])
        meta_dict = encode_dict_values_to_primitive(args["metadata"])

        self.assertEqual(raw_result[0].attributes, attr_dict)
        self.assertEqual(raw_result[0].metadata, meta_dict)

        # check bundle

    def test_9_get_records_by_filter_with_properties(self):
        """
        This test is to get a specific part of the graph via certain filter criteria


        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_9_get_records_by_filter_with_properties.svg
           :align: center
           :scale: 50 %

        *Get single node*

        The first part of the test is to try to get a single node based on a attribut

        *Input-Data*

        .. code-block:: json

            {
                "prov:type":"prov:Bundle"
            }


        *Output-Data*

        The output is a list of namedtuples wit the following structure: `list(tuple(attributes,metadata))`

        .. code-block:: json

            [
                [
                    {
                        "prov:type": "prov:Bundle"
                    },
                    {
                        "prov_type": "prov:Entity",
                        "identifier": "ex:bundle name",
                        "type_map": "{\\"date value\\": \\"xds:datetime\\", \\"int value\\": \\"int\\"}",
                        "namespaces": "{\\"ex\\": \\"http://example.com\\"}"
                    }
                ]
            ]


        *Get other nodes*

        The second part tests the other way to get all other node except the bundle node

        *Input-Data*

        The input is a set of attributes

        .. code-block:: json

            {
               "ex:dict value":{
                  "dict":"value"
               },
               "ex:double value":99.33,
               "ex:int value":99,
               "ex:individual attribute":"Some value",
               "ex:list value":[
                  "list",
                  "of",
                  "strings"
               ]
            }

        *Output-Data*

        The output is a list of namedtuple with attributes and metadata

        .. code-block:: json

            [
               [
                  {
                     "ex:dict value":"{'dict': 'value'}",
                     "ex:double value":99.33,
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ],
                     "ex:int value":99,
                     "ex:date value":"2005-06-01 13:33:00",
                     "ex:individual attribute":"Some value"
                  },
                  {
                     "identifier":"ex:TO NODE",
                     "type_map":"{'int value': 'int', 'date value': 'xds:datetime'}",
                     "namespaces":"{'ex': 'http://example.com', 'custom': 'http://custom.com'}",
                     "prov_type":"prov:Activity"
                  }
               ],
               [
                  {
                     "ex:dict value":"{'dict': 'value'}",
                     "ex:double value":99.33,
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ],
                     "ex:int value":99,
                     "ex:date value":"2005-06-01 13:33:00",
                     "ex:individual attribute":"Some value"
                  },
                  {
                     "identifier":"ex:FROM NODE",
                     "type_map":"{'int value': 'int', 'date value': 'xds:datetime'}",
                     "namespaces":"{'ex': 'http://example.com', 'custom': 'http://custom.com'}",
                     "prov_type":"prov:Activity"
                  }
               ],
               [
                  {
                     "ex:dict value":"{'dict': 'value'}",
                     "ex:double value":99.33,
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ],
                     "ex:int value":99,
                     "ex:date value":"2005-06-01 13:33:00",
                     "ex:individual attribute":"Some value"
                  },
                  {
                     "identifier":"ex:prov:example_node",
                     "type_map":"{'int value': 'int', 'date value': 'xds:datetime'}",
                     "namespaces":"{'ex': 'http://example.com', 'custom': 'http://custom.com'}",
                     "prov_type":"prov:Activity"
                  }
               ],
               [
                  {
                     "ex:dict value":"{'dict': 'value'}",
                     "ex:double value":99.33,
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ],
                     "ex:int value":99,
                     "ex:date value":"2005-06-01 13:33:00",
                     "ex:individual attribute":"Some value"
                  },
                  {
                     "identifier":"identifier for the relation",
                     "type_map":"{'int value': 'int', 'date value': 'xds:datetime'}",
                     "namespaces":"{'ex': 'http://example.com', 'custom': 'http://custom.com'}",
                     "prov_type":"prov:Mention"
                  }
               ]
            ]

        :return

        """
        self.clear_database()
        ids = insert_document_with_bundles(self.instance)

        # test get single node
        bundle_filter = dict()
        bundle_filter.update({PROV_TYPE: "prov:Bundle"})

        raw_bundle_nodes = self.instance.get_records_by_filter(attributes_dict=bundle_filter)

        self.assertIsNotNone(raw_bundle_nodes)
        self.assertIsInstance(raw_bundle_nodes, list)
        self.assertEqual(len(raw_bundle_nodes), 1)

        self.assertIsNotNone(raw_bundle_nodes[0].attributes)
        self.assertIsNotNone(raw_bundle_nodes[0].metadata)
        self.assertIsNotNone(raw_bundle_nodes[0].attributes[str(PROV_TYPE)])
        self.assertEqual(raw_bundle_nodes[0].attributes[str(PROV_TYPE)], bundle_filter[PROV_TYPE])

        # test get rest of the nodes
        # Use the same attributes as the insert_document_with_bundles() method
        node_filter = base_connector_record_parameter_example()["attributes"]

        # remove date value because this is individual for each node
        del node_filter["ex:date value"]

        raw_nodes = self.instance.get_records_by_filter(attributes_dict=node_filter)
        self.assertIsNotNone(raw_nodes)
        self.assertIsInstance(raw_nodes, list)
        self.assertEqual(len(raw_nodes),
                         4)  # 4 because the relation is also in the restults and 3 nodes, exclude the bundle node

        self.assertIsNotNone(raw_nodes[0].attributes)
        self.assertIsNotNone(raw_nodes[0].metadata)

        # check if the metadata of the record equals

        args = base_connector_record_parameter_example()

        # Remvoe date because it can't be the same
        del args["attributes"]["ex:date value"]
        del raw_nodes[2].attributes["ex:date value"]

        # Delete metdata prov:type because it can be different
        # del args["metadata"]["prov:type"]
        # del raw_nodes[0].metadata["prov:type"]

        attr_dict = encode_dict_values_to_primitive(args["attributes"])
        meta_dict = encode_dict_values_to_primitive(args["metadata"])

        del meta_dict[METADATA_KEY_IDENTIFIER]
        del raw_nodes[2].metadata[METADATA_KEY_IDENTIFIER]

        self.assertEqual(raw_nodes[2].attributes, attr_dict)
        self.assertEqual(raw_nodes[2].metadata, meta_dict)

    def test_10_get_records_by_filter_with_metadata(self):
        """
        Should test also the filter by metadata

        @todo implement test for filter by metadata


        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_10_get_records_by_filter_with_metadata.svg
           :align: center
           :scale: 50 %


        .. warning::
            This test is not implemented jet

        :return:
        """
        self.clear_database()
        args = base_connector_record_parameter_example()

        record_id = self.instance.save_element(args["attributes"], args["metadata"])

        raw_records = self.instance.get_records_by_filter()

        # Check return raw_records
        self.assertIsNotNone(raw_records)
        self.assertIsInstance(raw_records, list)
        self.assertEqual(len(raw_records), 1)
        self.assertIsInstance(raw_records[0].attributes, dict)
        self.assertIsInstance(raw_records[0].metadata, dict)

        # check if the metadata of the record equals
        attr_dict = encode_dict_values_to_primitive(args["attributes"])
        meta_dict = encode_dict_values_to_primitive(args["metadata"])

        self.assertEqual(raw_records[0].attributes, attr_dict)
        self.assertEqual(raw_records[0].metadata, meta_dict)

    def test_11_get_records_tail(self):
        """
        This test is to get the whole provenance from a starting point

        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_11_get_records_tail.svg
           :align: center
           :scale: 50 %

        **Input-Data**

        In this case we filter by metadata and by the identifier

        .. code-block:: json

            {
               "identifier":"ex:FROM NODE"
            }

        **Output-Data**

        The output is all connected nodes and there relations

        .. code-block:: json

            [
               [
                  {
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ],
                     "ex:double value":99.33,
                     "ex:int value":99,
                     "ex:individual attribute":"Some value",
                     "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                     "ex:dict value":{
                        "dict":"value"
                     }
                  },
                  {
                     "prov_type":"<QualifiedName: prov:Mention>",
                     "identifier":"identifier for the relation",
                     "type_map":{
                        "date value":"xds:datetime",
                        "int value":"int"
                     },
                     "namespaces":{
                        "custom":"http://custom.com",
                        "ex":"http://example.com"
                     }
                  }
               ],
               [
                  {
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ],
                     "ex:double value":99.33,
                     "ex:int value":99,
                     "ex:individual attribute":"Some value",
                     "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                     "ex:dict value":{
                        "dict":"value"
                     }
                  },
                  {
                     "prov_type":"<QualifiedName: prov:Activity>",
                     "identifier":"<QualifiedName: ex:TO NODE>",
                     "type_map":{
                        "date value":"xds:datetime",
                        "int value":"int"
                     },
                     "namespaces":{
                        "custom":"http://custom.com",
                        "ex":"http://example.com"
                     }
                  }
               ]
            ]

        """
        self.clear_database()
        ids = insert_document_with_bundles(self.instance)

        from_record = self.instance.get_record(ids["from_record_id"])
        to_record = self.instance.get_record(ids["to_record_id"])

        meta_filter = dict()
        meta_filter.update({METADATA_KEY_IDENTIFIER: from_record.metadata[METADATA_KEY_IDENTIFIER]})
        tail_records = self.instance.get_records_tail(metadata_dict=meta_filter)

        self.assertIsNotNone(tail_records)
        self.assertIsInstance(tail_records, list)
        self.assertEqual(len(tail_records), 2)  # 1 node and 1 relation
        self.assertIsNotNone(tail_records[0].attributes)
        self.assertIsNotNone(tail_records[0].metadata)
        self.assertIsInstance(tail_records[0].attributes, dict)
        self.assertIsInstance(tail_records[0].metadata, dict)

        if str(tail_records[0].metadata[METADATA_KEY_IDENTIFIER]) == str(
                to_record.metadata[METADATA_KEY_IDENTIFIER]):
            db_from_record = tail_records[0]
        elif str(tail_records[1].metadata[METADATA_KEY_IDENTIFIER]) == str(
                to_record.metadata[METADATA_KEY_IDENTIFIER]):
            db_from_record = tail_records[1]
        else:
            raise Exception("tail_record must contain the to_record identifier")

        tail_records_attr = encode_dict_values_to_primitive(db_from_record.attributes)
        tail_records_meta = encode_dict_values_to_primitive(db_from_record.metadata)
        self.assertEqual(tail_records_attr, to_record.attributes)
        self.assertEqual(tail_records_meta, to_record.metadata)

    def test_12_get_records_tail_recursive(self):
        """
        Test the same behavior as the `test_get_records_tail` test but with a recursive data structure

        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_12_get_records_tail_recursive.svg
           :align: center
           :scale: 50 %

        **Input-Data**

        .. code-block:: json

            {
               "identifier":"ex:FROM NODE"
            }

        **Output-Data**

        The output is all connected nodes and there relations

        .. code-block:: json

            [
               [
                  {
                     "ex:dict value":{
                        "dict":"value"
                     },
                     "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                     "ex:double value":99.33,
                     "ex:individual attribute":"Some value",
                     "ex:int value":99,
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ]
                  },
                  {
                     "namespaces":{
                        "custom":"http://custom.com",
                        "ex":"http://example.com"
                     },
                     "identifier":"identifier for the relation",
                     "prov_type":"<QualifiedName: prov:Mention>",
                     "type_map":{
                        "date value":"xds:datetime",
                        "int value":"int"
                     }
                  }
               ],
               [
                  {
                     "ex:dict value":{
                        "dict":"value"
                     },
                     "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                     "ex:double value":99.33,
                     "ex:individual attribute":"Some value",
                     "ex:int value":99,
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ]
                  },
                  {
                     "namespaces":{
                        "custom":"http://custom.com",
                        "ex":"http://example.com"
                     },
                     "identifier":"identifier for the relation",
                     "prov_type":"<QualifiedName: prov:Mention>",
                     "type_map":{
                        "date value":"xds:datetime",
                        "int value":"int"
                     }
                  }
               ],
               [
                  {
                     "ex:dict value":{
                        "dict":"value"
                     },
                     "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                     "ex:double value":99.33,
                     "ex:individual attribute":"Some value",
                     "ex:int value":99,
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ]
                  },
                  {
                     "namespaces":{
                        "custom":"http://custom.com",
                        "ex":"http://example.com"
                     },
                     "identifier":"<QualifiedName: ex:TO NODE>",
                     "prov_type":"<QualifiedName: prov:Activity>",
                     "type_map":{
                        "date value":"xds:datetime",
                        "int value":"int"
                     }
                  }
               ],
               [
                  {
                     "ex:dict value":{
                        "dict":"value"
                     },
                     "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                     "ex:double value":99.33,
                     "ex:individual attribute":"Some value",
                     "ex:int value":99,
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ]
                  },
                  {
                     "namespaces":{
                        "custom":"http://custom.com",
                        "ex":"http://example.com"
                     },
                     "identifier":"<QualifiedName: ex:second_TO NODE>",
                     "prov_type":"<QualifiedName: prov:Activity>",
                     "type_map":{
                        "date value":"xds:datetime",
                        "int value":"int"
                     }
                  }
               ],
               [
                  {
                     "ex:dict value":{
                        "dict":"value"
                     },
                     "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                     "ex:double value":99.33,
                     "ex:individual attribute":"Some value",
                     "ex:int value":99,
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ]
                  },
                  {
                     "namespaces":{
                        "custom":"http://custom.com",
                        "ex":"http://example.com"
                     },
                     "identifier":"identifier for the relation",
                     "prov_type":"<QualifiedName: prov:Mention>",
                     "type_map":{
                        "date value":"xds:datetime",
                        "int value":"int"
                     }
                  }
               ],
               [
                  {
                     "ex:dict value":{
                        "dict":"value"
                     },
                     "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                     "ex:double value":99.33,
                     "ex:individual attribute":"Some value",
                     "ex:int value":99,
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ]
                  },
                  {
                     "namespaces":{
                        "custom":"http://custom.com",
                        "ex":"http://example.com"
                     },
                     "identifier":"<QualifiedName: ex:FROM NODE>",
                     "prov_type":"<QualifiedName: prov:Activity>",
                     "type_map":{
                        "date value":"xds:datetime",
                        "int value":"int"
                     }
                  }
               ],
               [
                  {
                     "ex:dict value":{
                        "dict":"value"
                     },
                     "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                     "ex:double value":99.33,
                     "ex:individual attribute":"Some value",
                     "ex:int value":99,
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ]
                  },
                  {
                     "namespaces":{
                        "custom":"http://custom.com",
                        "ex":"http://example.com"
                     },
                     "identifier":"identifier for the relation",
                     "prov_type":"<QualifiedName: prov:Mention>",
                     "type_map":{
                        "date value":"xds:datetime",
                        "int value":"int"
                     }
                  }
               ],
               [
                  {
                     "ex:dict value":{
                        "dict":"value"
                     },
                     "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                     "ex:double value":99.33,
                     "ex:individual attribute":"Some value",
                     "ex:int value":99,
                     "ex:list value":[
                        "list",
                        "of",
                        "strings"
                     ]
                  },
                  {
                     "namespaces":{
                        "custom":"http://custom.com",
                        "ex":"http://example.com"
                     },
                     "identifier":"<QualifiedName: ex:second_FROM NODE>",
                     "prov_type":"<QualifiedName: prov:Activity>",
                     "type_map":{
                        "date value":"xds:datetime",
                        "int value":"int"
                     }
                  }
               ]
            ]



        """
        # def test_get_records_tail_recursive(self):

        self.clear_database()
        ids = insert_document_with_bundles(self.instance)
        ids2 = insert_document_with_bundles(self.instance, "second_")

        from_record = self.instance.get_record(ids["from_record_id"])
        to_record = self.instance.get_record(ids["to_record_id"])
        second_from_record = self.instance.get_record(ids2["from_record_id"])
        second_to_record = self.instance.get_record(ids2["to_record_id"])

        relation_params = base_connector_relation_parameter_example()

        self.instance.save_relation(to_record.metadata[METADATA_KEY_IDENTIFIER],
                                    second_from_record.metadata[METADATA_KEY_IDENTIFIER],
                                    relation_params["attributes"], relation_params["metadata"])
        self.instance.save_relation(second_from_record.metadata[METADATA_KEY_IDENTIFIER],
                                    from_record.metadata[METADATA_KEY_IDENTIFIER],
                                    relation_params["attributes"], relation_params["metadata"])

        meta_filter = dict()
        meta_filter.update({METADATA_KEY_IDENTIFIER: from_record.metadata[METADATA_KEY_IDENTIFIER]})
        tail_records = self.instance.get_records_tail(metadata_dict=meta_filter)

        self.assertIsNotNone(tail_records)
        self.assertIsInstance(tail_records, list)
        self.assertEqual(len(tail_records), 8)  # 4 Nodes and 4 connections
        self.assertIsNotNone(tail_records[0].attributes)
        self.assertIsNotNone(tail_records[0].metadata)
        self.assertIsInstance(tail_records[0].attributes, dict)
        self.assertIsInstance(tail_records[0].metadata, dict)

    def test_13_get_bundle_records(self):
        """
        The get_bundle function is to return the records (relation and nodes) for a bundle identifier

        Test the same behavior as the `test_get_records_tail` test but with a recursive data structure

        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_13_get_bundle_records.svg
           :align: center
           :scale: 50 %

        **Input-Data**

        .. code-block:: json

            {
               "identifier":"ex:FROM NODE"
            }

        **Output-Data**

        The output is all connected nodes and there relations

         .. warning::
            coming soon!


        """
        self.clear_database()
        #create relation in database
        doc = ProvDocument()

        from_record_args = base_connector_record_parameter_example()
        to_record_args = base_connector_record_parameter_example()
        relation_args = base_connector_relation_parameter_example()

        from_identifier = relation_args["from_node"]
        to_identifier = relation_args["to_node"]

        #set up bundle
        relation_args["metadata"][METADATA_KEY_PROV_TYPE] = PROV_RECORD_IDS_MAP["wasAssociatedWith"]
        relation_args["attributes"][PROV_TYPE] = doc.valid_qualified_name("prov:bundleAssociation")
        from_record_args["metadata"][METADATA_KEY_IDENTIFIER] = from_identifier
        to_record_args["metadata"][METADATA_KEY_IDENTIFIER] = to_identifier

        from_record_id = self.instance.save_element(from_record_args["attributes"], from_record_args["metadata"])  #
        to_record_id = self.instance.save_element(to_record_args["attributes"], to_record_args["metadata"])  #

        relation_id = self.instance.save_relation(to_identifier, from_identifier, relation_args["attributes"], relation_args["metadata"])

        # add one more
        args = base_connector_record_parameter_example()
        identifier   = doc.valid_qualified_name("prov:anotherNode")
        args["metadata"][METADATA_KEY_IDENTIFIER] =identifier

        relation_args["metadata"][METADATA_KEY_PROV_TYPE] = PROV_RECORD_IDS_MAP["wasAssociatedWith"]

        record_id = self.instance.save_element(args["attributes"], args["metadata"])
        relation_id = self.instance.save_relation(identifier, from_identifier, relation_args["attributes"], relation_args["metadata"])


        relation_args["metadata"][METADATA_KEY_PROV_TYPE] = PROV_RECORD_IDS_MAP["wasGeneratedBy"]
        relation_args["attributes"][PROV_TYPE] = PROV_RECORD_IDS_MAP["wasGeneratedBy"]
        relation_id = self.instance.save_relation(identifier, to_identifier, relation_args["attributes"], relation_args["metadata"])


        #Test the get
        result_records = self.instance.get_bundle_records("ex:Yoda")

        self.assertIsNotNone(result_records)
        self.assertIsInstance(result_records,list)
        self.assertTrue(len(result_records),2)# 2 x node (another node) ,  1 x relation (was informed by )

    ##Delete section ###
    def test_14_delete_by_filter(self):
        """
        Try to all records of the database

        :return:
        """
        self.clear_database()
        ids = insert_document_with_bundles(self.instance)
        result = self.instance.delete_records_by_filter()
        self.assertIsInstance(result,bool)
        self.assertTrue(result)

        with self.assertRaises(NotFoundException):
            from_record_id  = ids["from_record_id"]
            self.instance.get_record(from_record_id)


    def test_15_delete_by_filter_with_properties(self):
        """
        Try to delete by filter, same behavior as get_by_filter

        :return:
        """
        self.clear_database()
        ids = insert_document_with_bundles(self.instance)

        #Use the same attributes as the insert_document_with_bundles() method
        property_filter= base_connector_record_parameter_example()["attributes"]

        #remove date value because this is individual for each node
        del property_filter["ex:date value"]

        result = self.instance.delete_records_by_filter(property_filter)
        self.assertIsInstance(result,bool)
        self.assertTrue(result)

        with self.assertRaises(NotFoundException):
            from_record_id  = ids["from_record_id"]
            self.instance.get_record(from_record_id)

        with self.assertRaises(NotFoundException):
            to_record_id  = ids["to_record_id"]
            self.instance.get_record(to_record_id)


        #Bundle node should be there because not all attributes are match the filter
        bundle_id = self.instance.get_record(ids["bundle_id"])
        self.assertIsNotNone(bundle_id)

        #After the delete it should be only one node in the database
        raw_results = self.instance.get_records_by_filter(dict())
        self.assertIsNotNone(raw_results)
        self.assertIsInstance(raw_results, list)
        self.assertEqual(len(raw_results), 1)  #one relation and one node


    def test_16_delete_by_filter_with_metadata(self):
        """
        Try to delete by metadata, same behavior as get_by_metadata
        :return:
        """
        self.clear_database()
        ids = insert_document_with_bundles(self.instance)

        from_record_id = ids["from_record_id"]

        from_record = self.instance.get_record(from_record_id)

        # Use the same attributes as the insert_document_with_bundles() method
        property_filter = dict()
        #Get identifier from record and add the identifer to the filter
        property_filter.update({METADATA_KEY_IDENTIFIER: from_record.metadata[METADATA_KEY_IDENTIFIER]})


        #Try to delete only the from node
        result = self.instance.delete_records_by_filter(metadata_dict=property_filter)
        self.assertIsInstance(result, bool)
        self.assertTrue(result)

        with self.assertRaises(NotFoundException):
            from_record_id = ids["from_record_id"]
            self.instance.get_record(from_record_id)

        # After the delete it should be one less in the db
        raw_results = self.instance.get_records_by_filter()
        self.assertIsNotNone(raw_results)
        self.assertIsInstance(raw_results, list)
        self.assertEqual(len(raw_results), 3)

    def test_17_delete_record(self):
        """
        Delete a single record based on the id

        :return:
        """
        self.clear_database()
        ids = insert_document_with_bundles(self.instance)
        from_record_id = ids["from_record_id"]
        result = self.instance.delete_record(from_record_id)
        self.assertIsInstance(result, bool)
        self.assertTrue(result)

        with self.assertRaises(NotFoundException):
            self.instance.get_record(from_record_id)

    def test_18_delete_relation(self):
        """
        Delete a singe relation based on the relation id

        :return:
        """
        self.clear_database()
        ids = insert_document_with_bundles(self.instance)
        relation_id = ids["relation_id"]
        result = self.instance.delete_relation(relation_id)
        self.assertIsInstance(result, bool)
        self.assertTrue(result)

        with self.assertRaises(NotFoundException):
            self.instance.get_relation(relation_id)

    ### Merge documents ###

    def test_19_merge_record(self):
        """
        This function test the merge abbility of your adapter.


        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_19_merge_record.svg
           :align: center
           :scale: 50 %

        **Input-Data**

        We try to create the node twice, with the following data

        .. code-block:: json

            {
               "metadata":{
                  "prov_type":"<QualifiedName: prov:Activity>",
                  "namespaces":{
                     "ex":"http://example.com",
                     "custom":"http://custom.com"
                  },
                  "identifier":"<QualifiedName: ex:Yoda>",
                  "type_map":{
                     "int value":"int",
                     "date value":"xds:datetime"
                  }
               },
               "attributes":{
                  "ex:int value":99,
                  "ex:individual attribute":"Some value",
                  "ex:list value":[
                     "list",
                     "of",
                     "strings"
                  ],
                  "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
                  "ex:dict value":{
                     "dict":"value"
                  },
                  "ex:double value":99.33
               }
            }

        **Output-Data**

        The output is one entry with no change of the data

        .. code-block:: json

            [
               {
                  "ex:int value":99,
                  "ex:individual attribute":"Some value",
                  "ex:list value":[
                     "list",
                     "of",
                     "strings"
                  ],
                  "ex:date value":"2005-06-01 13:33:00",
                  "ex:dict value":"{\"dict\": \"value\"}",
                  "ex:double value":99.33
               },
               {
                  "prov_type":"prov:Activity",
                  "namespaces":"{\"ex\": \"http://example.com\", \"custom\": \"http://custom.com\"}",
                  "identifier":"ex:Yoda",
                  "type_map":"{\"int value\": \"int\", \"date value\": \"xds:datetime\"}"
               }
            ]


        """
        self.clear_database()
        example = base_connector_merge_example()
        #Skip test if this merge mode is not supported

        #save record test
        rec_id1 = self.instance.save_element(example.from_node["attributes"], example.from_node["metadata"])
        rec_id2 = self.instance.save_element(example.from_node["attributes"], example.from_node["metadata"])

        self.assertEqual(rec_id1,rec_id2)

        # Test merge result of save record

        db_record = self.instance.get_record(rec_id2) # The id's are equal so it makes no difference

        attr = encode_dict_values_to_primitive(example.from_node["attributes"])
        meta = encode_dict_values_to_primitive(example.from_node["metadata"])


        db_meta = encode_adapter_result_to_excpect(db_record.metadata).copy()
        #transform string into dict to compare result

        meta.update({METADATA_KEY_NAMESPACES: json.loads(meta[METADATA_KEY_NAMESPACES])})
        meta.update({METADATA_KEY_TYPE_MAP: json.loads(meta[METADATA_KEY_TYPE_MAP])})

        db_meta.update({METADATA_KEY_NAMESPACES: json.loads(db_meta[METADATA_KEY_NAMESPACES])})
        db_meta.update({METADATA_KEY_TYPE_MAP: json.loads(db_meta[METADATA_KEY_TYPE_MAP])})
        self.assertEqual(attr,db_record.attributes)


        self.assertEqual(meta,db_meta)
        prim = primer_example()
        self.assertEqual(len(prim.get_records()),len(prim.unified().get_records()))

    def test_20_merge_record_complex(self):
        """
        In this example we test if we merge different attributes into one node


        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_20_merge_record_complex.svg
           :align: center
           :scale: 50 %

        **Input-Data**

        This is the attributes used to create the entry

        .. code-block:: json

            {
               "ex:individual attribute":"Some value",
               "ex:dict value":{
                  "dict":"value"
               },
               "ex:double value":99.33,
               "ex:list value":[
                  "list",
                  "of",
                  "strings"
               ],
               "ex:int value":99,
               "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)"
            }

        This are the attributes to alter the existing node

        .. code-block:: json

            {
                "ex:a other attribute":true
            }

        **Output-Data**

        The output is one entry with the additional attribute

        .. code-block:: json

            [
               {
                  "ex:individual attribute":"Some value",
                  "ex:dict value":"{\"dict\": \"value\"}",
                  "ex:double value":99.33,
                  "ex:list value":[
                     "list",
                     "of",
                     "strings"
                  ],
                  "ex:int value":99,
                  "ex:date value":"2005-06-01 13:33:00",
                  "ex:a other attribute":true
               },
               {
                  "type_map":"{\"date value\": \"xds:datetime\", \"int value\": \"int\"}",
                  "identifier":"ex:Yoda",
                  "prov_type":"prov:Activity",
                  "namespaces":"{\"ex\": \"http://example.com\", \"custom\": \"http://custom.com\"}"
               }
            ]


        """
        self.clear_database()
        example = base_connector_merge_example()
        # Skip test if this merge mode is not supported

        # Save record with different attributes
        rec_id1 = self.instance.save_element(example.from_node["attributes"], example.from_node["metadata"])

        attr_modified = dict()
        attr_modified.update({"ex:a other attribute": True})
        metadata_modified = example.from_node["metadata"].copy()

        rec_id2 = self.instance.save_element(attr_modified, metadata_modified)

        self.assertEqual(rec_id1, rec_id2)

        # Test merge result of save record

        db_record = self.instance.get_record(rec_id2)  # The id's are equal so it makes no difference

        attr = encode_dict_values_to_primitive(example.from_node["attributes"])
        attr.update(attr_modified) # add additional attr to check dict
        meta = encode_dict_values_to_primitive(example.from_node["metadata"])

        db_meta = encode_adapter_result_to_excpect(db_record.metadata).copy()

        meta.update({METADATA_KEY_NAMESPACES: json.loads(meta[METADATA_KEY_NAMESPACES])})
        meta.update({METADATA_KEY_TYPE_MAP: json.loads(meta[METADATA_KEY_TYPE_MAP])})

        db_meta.update({METADATA_KEY_NAMESPACES: json.loads(db_meta[METADATA_KEY_NAMESPACES])})
        db_meta.update({METADATA_KEY_TYPE_MAP: json.loads(db_meta[METADATA_KEY_TYPE_MAP])})


        self.assertEqual(attr, db_record.attributes)
        self.assertEqual(meta, db_meta)

    def test_21_merge_record_complex_fail(self):
        """
         In this example we test if we merge different attributes into one node

        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_21_merge_record_complex_fail.svg
           :align: center
           :scale: 50 %

         **Input-Data**

         This is the attributes used to create the entry

         .. code-block:: json

             {
               "ex:list value":[
                  "list",
                  "of",
                  "strings"
               ],
               "ex:double value":99.33,
               "ex:date value":"datetime.datetime(2005, 6, 1, 13, 33)",
               "ex:dict value":{
                  "dict":"value"
               },
               "ex:int value":99,
               "ex:individual attribute":"Some value"
            }

         Try to *override* the existing attribute

         .. code-block:: json

            {
                "ex:int value": 1
            }

         **Output-Data**

         Should throw an MergeException

         """
        self.clear_database()
        example = base_connector_merge_example()
        # Skip test if this merge mode is not supported

        # Save record with different attributes
        rec_id1 = self.instance.save_element(example.from_node["attributes"], example.from_node["metadata"])

        attr_modified = dict()
        attr_modified.update({"ex:int value": 1}) # set int value to 1 instead of 99, should throw exception
        metadata_modified = example.from_node["metadata"].copy()

        #should raise exception because otherwise the attribute would be overridden
        with self.assertRaises(MergeException):
            self.instance.save_element(attr_modified, metadata_modified)

    def test_22_merge_record_metadata(self):
        """
        This test try to merge the metadata.
        This is important if you add some new attributes that uses other namespaces, so you need to merge the namespaces
        Same behavior for the type_map

        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_22_merge_record_metadata.svg
           :align: center
           :scale: 50 %

        **Input-Data**

        The metadata for the initial record:

        .. code-block:: json

             {
               "type_map":{
                  "date value":"xds:datetime",
                  "int value":"int"
               },
               "prov_type":"<QualifiedName: prov:Activity>",
               "namespaces":{
                  "custom":"http://custom.com",
                  "ex":"http://example.com"
               },
               "identifier":"<QualifiedName: ex:Yoda>"
            }


        Try to add a record with some modified namespaces

        .. code-block:: json

            {
               "namespaces":{
                  "custom":"http://custom.com",
                  "ex":"http://example.com"
               },
               "prov_type":"<QualifiedName: prov:Activity>",
               "identifier":"<QualifiedName: ex:Yoda>",
               "type_map":{
                  "custom_attr_1":"xds:some_value"
               }
            }


        **Output-Data**

        The output is the merged result of the type map

        .. code-block:: json

            [
               {
                  "ex:individual attribute":"Some value",
                  "ex:list value":[
                     "list",
                     "of",
                     "strings"
                  ],
                  "ex:int value":99,
                  "ex:date value":"2005-06-01 13:33:00",
                  "ex:dict value":"{\"dict\": \"value\"}",
                  "ex:double value":99.33
               },
               {
                  "prov_type":"prov:Activity",
                  "type_map":"{\"custom_attr_1\": \"xds:some_value\", \"date value\": \"xds:datetime\", \"int value\": \"int\"}",
                  "identifier":"ex:Yoda",
                  "namespaces":"{\"custom\": \"http://custom.com\", \"ex\": \"http://example.com\"}"
               }
            ]

        """
        self.clear_database()
        example = base_connector_merge_example()
        # Skip test if this merge mode is not supported

        # Save record with different attributes
        rec_id1 = self.instance.save_element(example.from_node["attributes"], example.from_node["metadata"])


        metadata_modified = example.from_node["metadata"].copy()
        metadata_modified.update({METADATA_KEY_TYPE_MAP: {"custom_attr_1": "xds:some_value"}})

        rec_id2 = self.instance.save_element(example.from_node["attributes"], metadata_modified)


        self.assertEqual(rec_id1, rec_id2)

        db_record = self.instance.get_record(rec_id1)
        # try to merge type_map in the database

        attr = encode_dict_values_to_primitive(example.relation["attributes"])
        meta = encode_dict_values_to_primitive(example.relation["metadata"])
        meta_custom = encode_dict_values_to_primitive(metadata_modified)

        # check namespaces
        db_record_namespaces = db_record.metadata[METADATA_KEY_NAMESPACES]




        if type(db_record_namespaces) is list:
            self.assertIsInstance(db_record_namespaces, list)
            self.assertTrue(len(db_record_namespaces), 2)
            self.assertEqual(meta[METADATA_KEY_NAMESPACES], db_record_namespaces[0])
            self.assertEqual(meta_custom[METADATA_KEY_NAMESPACES], db_record_namespaces[1])
        else:
            self.assertIsInstance(db_record_namespaces, str)
            decoded_map = json.loads(db_record_namespaces)
            expected_map = json.loads(meta[METADATA_KEY_NAMESPACES])
            expected_map.update(json.loads(meta_custom[METADATA_KEY_NAMESPACES]))
            self.assertEqual(decoded_map, expected_map)


        # check type_map
        db_record_type_map = db_record.metadata[METADATA_KEY_TYPE_MAP]

        if type(db_record_type_map) is list:
            self.assertIsInstance(db_record_type_map, list)
            self.assertTrue(len(db_record_type_map), 2)
            self.assertEqual(meta[METADATA_KEY_TYPE_MAP], db_record_type_map[0])
            self.assertEqual(meta_custom[METADATA_KEY_TYPE_MAP], db_record_type_map[1])
        else:
            self.assertIsInstance(db_record_type_map, str)
            decoded_map = json.loads(db_record_type_map)
            expected_map = json.loads(meta[METADATA_KEY_TYPE_MAP])
            expected_map.update(json.loads(meta_custom[METADATA_KEY_TYPE_MAP]))
            self.assertEqual(decoded_map, expected_map)



    def test_23_merge_relation(self):
        """
        Merge a relation is pretty similar to merge records.
        The big difference is the different rules for uniques

        **Graph-Strucutre**

        .. figure:: https://cdn.rawgit.com/dlr-sc/prov-db-connector/master/docs/_images/test_cases/test_23_merge_relation.svg
           :align: center
           :scale: 50 %

        A relation is unique if:

        - The relation type is the same
        - all other formal attributes (see SimpleDbAdapter) are the same

        otherwise it is not the same relation.


        """
        self.clear_database()
        example = base_connector_merge_example()
        prim = primer_example()
        self.assertEqual(len(prim.get_records()), len(prim.unified().get_records()))
        #save relation test
        self.instance.save_element(example.from_node["attributes"], example.from_node["metadata"])
        self.instance.save_element(example.to_node["attributes"], example.to_node["metadata"])

        from_label = example.from_node["metadata"][METADATA_KEY_IDENTIFIER]
        to_label = example.to_node["metadata"][METADATA_KEY_IDENTIFIER]

        rel_id1 = self.instance.save_relation(from_label,to_label,example.relation["attributes"], example.relation["metadata"])
        rel_id2 = self.instance.save_relation(from_label,to_label,example.relation["attributes"], example.relation["metadata"])

        self.assertEqual(rel_id1,rel_id2)

        # Test merge result of save relation

        db_record = self.instance.get_relation(rel_id2)  # The id's are equal so it makes no difference

        attr = encode_dict_values_to_primitive(example.relation["attributes"])
        meta = encode_dict_values_to_primitive(example.relation["metadata"])

        db_meta = encode_adapter_result_to_excpect(db_record.metadata).copy()

        meta.update({METADATA_KEY_NAMESPACES: json.loads(meta[METADATA_KEY_NAMESPACES])})
        meta.update({METADATA_KEY_TYPE_MAP: json.loads(meta[METADATA_KEY_TYPE_MAP])})

        db_meta.update({METADATA_KEY_NAMESPACES: json.loads(db_meta[METADATA_KEY_NAMESPACES])})
        db_meta.update({METADATA_KEY_TYPE_MAP: json.loads(db_meta[METADATA_KEY_TYPE_MAP])})


        self.assertEqual(attr, db_record.attributes)
        self.assertEqual(meta, db_meta)


    def test_24_merge_relation_complex(self):
        """
        Same behavior as the merge_node test

        """
        self.clear_database()
        example = base_connector_merge_example()
        prim = primer_example()
        self.assertEqual(len(prim.get_records()), len(prim.unified().get_records()))
        #save relation test
        self.instance.save_element(example.from_node["attributes"], example.from_node["metadata"])
        self.instance.save_element(example.to_node["attributes"], example.to_node["metadata"])

        from_label = example.from_node["metadata"][METADATA_KEY_IDENTIFIER]
        to_label = example.to_node["metadata"][METADATA_KEY_IDENTIFIER]

        custom_attributes = dict()
        custom_attributes.update({"My custom value": "value"})

        rel_id1 = self.instance.save_relation(from_label,to_label,example.relation["attributes"], example.relation["metadata"])
        rel_id2 = self.instance.save_relation(from_label,to_label,custom_attributes, example.relation["metadata"])

        self.assertEqual(rel_id1,rel_id2)

        # Test merge result of save relation

        db_record = self.instance.get_relation(rel_id2)  # The id's are equal so it makes no difference

        example.relation["attributes"].update(custom_attributes)
        attr = encode_dict_values_to_primitive(example.relation["attributes"])
        meta = encode_dict_values_to_primitive(example.relation["metadata"])

        db_meta = encode_adapter_result_to_excpect(db_record.metadata).copy()

        meta.update({METADATA_KEY_NAMESPACES: json.loads(meta[METADATA_KEY_NAMESPACES])})
        meta.update({METADATA_KEY_TYPE_MAP: json.loads(meta[METADATA_KEY_TYPE_MAP])})

        db_meta.update({METADATA_KEY_NAMESPACES: json.loads(db_meta[METADATA_KEY_NAMESPACES])})
        db_meta.update({METADATA_KEY_TYPE_MAP: json.loads(db_meta[METADATA_KEY_TYPE_MAP])})


        self.assertEqual(attr, db_record.attributes)
        self.assertEqual(meta, db_meta)


    def test_25_merge_relation_complex_fail(self):
        """
        Same behavior as the merge_node_fail test

        """
        self.clear_database()
        example = base_connector_merge_example()
        prim = primer_example()
        self.assertEqual(len(prim.get_records()), len(prim.unified().get_records()))
        # save relation test
        self.instance.save_element(example.from_node["attributes"], example.from_node["metadata"])
        self.instance.save_element(example.to_node["attributes"], example.to_node["metadata"])

        from_label = example.from_node["metadata"][METADATA_KEY_IDENTIFIER]
        to_label = example.to_node["metadata"][METADATA_KEY_IDENTIFIER]

        #try ot override a propertie in the database
        custom_attributes = dict()
        custom_attributes.update({"ex:int value": 1})

        rel_id1 = self.instance.save_relation(from_label, to_label, example.relation["attributes"],
                                              example.relation["metadata"])

        with self.assertRaises(MergeException):
            rel_id2 = self.instance.save_relation(from_label, to_label, custom_attributes, example.relation["metadata"])


    def test_26_merge_relation_metadata(self):
        """
        Same as the merge_record_metadata

        """
        self.clear_database()
        example = base_connector_merge_example()
        prim = primer_example()
        self.assertEqual(len(prim.get_records()), len(prim.unified().get_records()))
        # save relation test
        self.instance.save_element(example.from_node["attributes"], example.from_node["metadata"])
        self.instance.save_element(example.to_node["attributes"], example.to_node["metadata"])

        from_label = example.from_node["metadata"][METADATA_KEY_IDENTIFIER]
        to_label = example.to_node["metadata"][METADATA_KEY_IDENTIFIER]

        # try to merge type_map in the database
        custom_metadata = example.relation["metadata"].copy()
        custom_metadata.update({METADATA_KEY_TYPE_MAP: {"custom_attr_1": "xds:some_value"}})

        self.instance.save_relation(from_label, to_label, example.relation["attributes"],
                                              example.relation["metadata"])

        rel_id2 = self.instance.save_relation(from_label, to_label, example.relation["attributes"], custom_metadata)


        db_record = self.instance.get_relation(rel_id2)  # The id's are equal so it makes no difference


        attr = encode_dict_values_to_primitive(example.relation["attributes"])
        meta = encode_dict_values_to_primitive(example.relation["metadata"])
        meta_custom = encode_dict_values_to_primitive(custom_metadata)


        self.assertEqual(attr, db_record.attributes)

        # check namespaces
        db_record_namespaces = db_record.metadata[METADATA_KEY_NAMESPACES]

        if type(db_record_namespaces) is list:
            self.assertIsInstance(db_record_namespaces, list)
            self.assertTrue(len(db_record_namespaces), 2)
            self.assertEqual(meta[METADATA_KEY_NAMESPACES], db_record_namespaces[0])
            self.assertEqual(meta_custom[METADATA_KEY_NAMESPACES], db_record_namespaces[1])
        else:
            self.assertIsInstance(db_record_namespaces, str)
            decoded_map = json.loads(db_record_namespaces)
            expected_map = json.loads(meta[METADATA_KEY_NAMESPACES])
            expected_map.update(json.loads(meta_custom[METADATA_KEY_NAMESPACES]))
            self.assertEqual(decoded_map, expected_map)

        # check type_map
        db_record_type_map = db_record.metadata[METADATA_KEY_TYPE_MAP]

        if type(db_record_type_map) is list:
            self.assertIsInstance(db_record_type_map, list)
            self.assertTrue(len(db_record_type_map), 2)
            self.assertEqual(meta[METADATA_KEY_TYPE_MAP], db_record_type_map[0])
            self.assertEqual(meta_custom[METADATA_KEY_TYPE_MAP], db_record_type_map[1])
        else:
            self.assertIsInstance(db_record_type_map, str)
            decoded_map = json.loads(db_record_type_map)
            expected_map = json.loads(meta[METADATA_KEY_TYPE_MAP])
            expected_map.update(json.loads(meta_custom[METADATA_KEY_TYPE_MAP]))
            self.assertEqual(decoded_map, expected_map)



    #Rest of the tests
    def test_27_save_bundle(self):
        """
        .. warning:
            Don't start with this function

        :return:
        """
        self.clear_database()
        args = base_connector_bundle_parameter_example()
        id = self.instance.save_element(args["attributes"], args["metadata"])
        self.assertIsNotNone(id)
        self.assertIs(type(id), str, "id should be a string ")

class BaseConnectorTests(unittest.TestCase):
    """
    This class is only to test that the BaseConnector is alright

    """
    def test_instance_abstract_class(self):
        """
        Test that the BaseAdapter is abstract
        """
        base = BaseAdapter()
        with self.assertRaises(NotImplementedError):
            base.connect(None)


import unittest


class ValidatorTests(unittest.TestCase):
    """
    Test the validator class

    """
    def setUp(self):
        """
        Setup validator
        """
        pass

    def tearDown(self):
        """
        Delete validator
        """
        pass




import logging
from uuid import uuid4

from prov.constants import PROV_ASSOCIATION, PROV_TYPE, PROV_MENTION
from provdbconnector.db_adapters.baseadapter import BaseAdapter, DbRecord, DbRelation, METADATA_KEY_IDENTIFIER, \
    METADATA_KEY_PROV_TYPE
from provdbconnector.exceptions.database import InvalidOptionsException, NotFoundException
from provdbconnector.utils.serializer import encode_dict_values_to_primitive, split_into_formal_and_other_attributes, \
    merge_record

log = logging.getLogger(__name__)


class SimpleInMemoryAdapter(BaseAdapter):
    """
    The simple in memory adapter is a reference implementation for a database adapter to save prov information
    into a graph database


    For exmaple to use the simple db_adapter use the following script

    .. literalinclude:: ../examples/simple_example.py
            :linenos:
            :language: python

    """
    all_nodes = dict()  # separate dict for records only (to get them by id)
    """
    Contains all nodes
    """
    all_relations = dict()
    """
    Contains all relation according to the following structure
    `(start_identifier, (end_identifier,attributes, metadata))``
    """

    def __init__(self, *args):
        """
        Init the adapter without any params
        :param args:
        """
        super(SimpleInMemoryAdapter, self).__init__()
        pass

    def connect(self, authentication_info):
        """
        This function setups your database connection (auth / service discover)

        :param authentication_info: The info to connect to the db
        :type authentication_info: dict or None
        :return: The result of the connection attempt
        :rtype: Bool
        """

        if authentication_info is not None:
            raise InvalidOptionsException()

        return True

    def save_element(self, attributes, metadata):
        """
        Store a single node in the database and if necessary and possible merge the node

        :param attributes: The actual provenance data
        :type attributes: dict
        :param metadata: Some metadata that are not PROV-O related
        :type metadata: dict
        :return: id of the record
        :rtype: str
        """

        # because it is in memory, we should copy the dicts to prevent others from modify the data
        attributes = attributes.copy()
        metadata = metadata.copy()

        # save all record information and return record id as string

        identifier = metadata[METADATA_KEY_IDENTIFIER]
        if str(identifier) in self.all_nodes:
            # try to merge nodes
            (old_attributes, old_metadata) = self.all_nodes[str(identifier)]
            (merged_attributes, merged_metadata) = merge_record(old_attributes, old_metadata, attributes, metadata)

            self.all_nodes.update({str(identifier): (merged_attributes, merged_metadata)})

        else:

            # encode your variables, based on your database architecture
            # (in this case it is not really necessary but for demonstration propose I saved the encoded vars )
            # attr = encode_dict_values_to_primitive(attributes)
            # meta = encode_dict_values_to_primitive(metadata)

            self.all_nodes.update({str(identifier): (attributes, metadata)})

        return str(identifier)

    def save_relation(self, from_node, to_node, attributes, metadata):
        """
        Store a relation between 2 nodes in the database.
        Merge also the relation if necessary and possible

        :param from_node: The identifier for the start node
        :type from_node: prov.model.Identifier
        :param to_node: The identifier for the end node
        :type to_node: prov.model.Identifier
        :param attributes: The actual provenance data
        :type attributes: dict
        :param metadata: Some metadata that are not PROV-O related
        :type metadata: dict
        :return: The id of the relation
        :rtype: str
        """

        # save all relation information and return the relation id as string

        # because it is in memory, we should copy the dicts to prevent others from modify the data
        attributes = attributes.copy()
        metadata = metadata.copy()

        # add dict if it is the first relation
        if str(from_node) not in self.all_relations:
            self.all_relations.update({str(from_node): dict()})

        # ===============
        # MERGE RELATION
        # ===============
        new_relation_formal_attributes = split_into_formal_and_other_attributes(attributes, metadata)

        # check that the from node already has some relations
        if str(from_node) in self.all_relations:

            # for each relation with the origin "from_node"
            for (relation_id, (to_identifier, old_attributes, old_metadata)) in self.all_relations[
                str(from_node)].items():
                # check if connection is to the same identifier
                if str(to_node) == to_identifier and metadata[METADATA_KEY_PROV_TYPE] == old_metadata[
                    METADATA_KEY_PROV_TYPE]:
                    # okay got potential duplicate... lets check the formal attributes
                    old_relation_formal_attributes = split_into_formal_and_other_attributes(old_attributes,
                                                                                            old_metadata)

                    if old_relation_formal_attributes.formal == new_relation_formal_attributes.formal:
                        # got duplicate
                        (merged_attributes, merged_metadata) = merge_record(old_attributes, old_metadata, attributes,
                                                                            metadata)
                        self.all_relations[str(from_node)].update(
                            {relation_id: (to_identifier, merged_attributes, merged_metadata)})
                        return relation_id

        # ===============
        # CREATE NEW RELATION
        # ===============

        id = str(uuid4())

        relations = self.all_relations[str(from_node)]
        relations.update({id: (str(to_node), attributes, metadata)})

        return id

    def get_record(self, record_id):
        """
        Get a ProvDocument from the database based on the document id

        :param record_id: The id of the node
        :type record_id: str
        :return: A named tuple with (attributes, metadata)
        :rtype: DbRecord
        """
        if record_id not in self.all_nodes:
            raise NotFoundException()

        (attributes, metadata) = self.all_nodes.get(record_id)
        attributes = encode_dict_values_to_primitive(attributes)
        metadata = encode_dict_values_to_primitive(metadata)
        db_record = DbRecord(attributes, metadata)

        return db_record

    def get_relation(self, relation_id):
        """
        Return the relation behind the relation_id

        :param relation_id: The id of the relation
        :type relation_id: str
        :return: The namedtuple with (attributes, metadata)
        :rtype: DbRelation
        """

        for (from_uri, relations) in self.all_relations.items():
            if relation_id in relations:
                (to_uri, attributes, metadata) = relations[relation_id]

                attributes = encode_dict_values_to_primitive(attributes)
                metadata = encode_dict_values_to_primitive(metadata)

                return DbRelation(attributes, metadata)
        raise NotFoundException("could't find the relation with id {}".format(relation_id))

    def get_records_by_filter(self, attributes_dict=None, metadata_dict=None):
        """
        Filter all nodes based on the provided attributes and metadata dict
        The filter is currently defined as follows:

        - The filter is only applied to the start node
        - All connections from the start node are also included in the result set

        :param attributes_dict: A filter dict with a conjunction of all values in the attributes_dict and metadata_dict
        :type attributes_dict: dict
        :param metadata_dict: A filter for the metadata with a conjunction of all values (also in the attributes_dict )
        :type metadata_dict: dict
        :return: The list of matching relations and nodes
        :rtype: List(DbRecord or Dbrelation)
        """
        if attributes_dict is None:
            attributes_dict = dict()
        if metadata_dict is None:
            metadata_dict = dict()

        return_records = list()
        return_keys = set()
        properties_filter_dict = encode_dict_values_to_primitive(attributes_dict.copy())
        metadata_filter_dict = encode_dict_values_to_primitive(metadata_dict.copy())
        for (identifier, (attributes, metadata)) in self.all_nodes.items():

            if self._check_attribute_metadata_filter(attributes_filter=properties_filter_dict,
                                                     metadata_filter=metadata_filter_dict,
                                                     metadata=metadata,
                                                     attributes=attributes):

                meta_encoded = encode_dict_values_to_primitive(metadata)
                attr_encoded = encode_dict_values_to_primitive(attributes)

                return_records.append(DbRecord(attr_encoded, meta_encoded))
                return_keys.add(identifier)

            else:
                # not match so don't add
                pass

        # get relations
        for (from_id, relations) in self.all_relations.items():

            if from_id in return_keys:

                for (relation_id, (to_id, attributes, metadata)) in relations.items():
                    if to_id in return_keys:
                        attributes = encode_dict_values_to_primitive(attributes)
                        metadata = encode_dict_values_to_primitive(metadata)

                        return_records.append(DbRelation(attributes, metadata))

        return return_records

    def get_records_tail(self, attributes_dict=None, metadata_dict=None, depth=None):
        """
        Return the provenance based on a filter combination.
        The filter dicts are only relevant for the start nodes.
        They describe the params to get the start nodes (for example a filter for a specific identifier ) and from there
        we want all connected nodes

        :param attributes_dict: A filter dict with a conjunction of all values in the attributes_dict and metadata_dict
        :type attributes_dict: dict
        :param metadata_dict: A filter for the metadata with a conjunction of all values (also in the attributes_dict )
        :type metadata_dict: dict
        :param depth: The level of detail, default to infinite
        :type depth: int
        :return: A list of DbRelations and DbRecords
        :rtype: list(DbRelation or DbRecord)
        """
        # only transform dict into list
        return list(self._get_records_tail_internal(attributes_dict, metadata_dict).values())

    def _get_records_tail_internal(self, attributes_dict=None, metadata_dict=None, max_depth=None, current_depth=0,
                                   result_records=None):
        """
        Internal function, because we need to call it recursive. This function is only neccecary in this in memory adapter.
        It's your choice how you implement the get_records_tail function

        :param attributes_dict: A filter dict with a conjunction of all values in the attributes_dict and metadata_dict
        :type attributes_dict: dict
        :param metadata_dict: A filter for the metadata with a conjunction of all values (also in the attributes_dict )
        :type metadata_dict: dict
        :param max_depth:
        :param current_depth:
        :param result_records:
        :return:
        """

        if attributes_dict is None:
            attributes_dict = dict()
        if metadata_dict is None:
            metadata_dict = dict()

        origin_records = self.get_records_by_filter(attributes_dict, metadata_dict)

        if result_records is None:
            result_records = dict()
        if current_depth == max_depth:
            return dict()
        for record in origin_records:
            from_identifier = record.metadata[METADATA_KEY_IDENTIFIER]
            if from_identifier in self.all_relations:
                for (relation_id, (to_identifier, attributes, metadata)) in self.all_relations[from_identifier].items():
                    # find to node
                    if to_identifier not in result_records:
                        (to_attributes, to_metadata) = self.all_nodes[to_identifier]
                        result_records.update({to_identifier: DbRecord(to_attributes, to_metadata)})

                        # add other nodes recursive
                        result_records.update(
                            self._get_records_tail_internal(metadata_dict={METADATA_KEY_IDENTIFIER: to_identifier},
                                                            current_depth=current_depth + 1,
                                                            result_records=result_records))

                    # add relation to result
                    result_records.update({relation_id: DbRelation(attributes, metadata)})

        return result_records

    def get_bundle_records(self, bundle_identifier):
        """
        Get the records for a specific bundle identifier

        This include all nodes that have a relation of the prov:type = prov:bundleAssociation and also
        all relation where the start and end node are in the bundle.
        Also you should add the prov mentionOf relation where the start node is in the bundle.
        See https://www.w3.org/TR/prov-links/

        :param bundle_identifier: The identifier of the bundle
        :type bundle_identifier: prov.model.Identifier
        :return: The list with the bundle nodes and all connections where the start node and end node in the bundle.
        :rtype: list(DbRelation or DbRecord )
        """
        bundle_records = dict()

        # get all nodes for the bundle
        for (from_identifier, relations) in self.all_relations.items():
            # search in all relations for a relation that points to the bundle and also fulfill the other constraints
            for (relation_id, (to_identifier, attributes, metadata)) in relations.items():

                if to_identifier == str(bundle_identifier):
                    # got potential bundle association, check prov:type to be sure

                    if metadata[METADATA_KEY_PROV_TYPE] == PROV_ASSOCIATION and str(
                            attributes[PROV_TYPE]) == "prov:bundleAssociation":
                        (attributes, metadata) = self.all_nodes[from_identifier]
                        bundle_records.update({from_identifier: DbRecord(attributes, metadata)})

        # search for all relations between the bundle nodes
        for from_identifier in bundle_records.copy().keys():
            if from_identifier in self.all_relations:
                for (relation_id, (to_identifier, attributes, metadata)) in self.all_relations[from_identifier].items():

                    # If the target of the relation is also in the bundle the relation belongs to the bundle
                    if to_identifier in bundle_records:
                        bundle_records.update({relation_id: DbRelation(attributes, metadata)})

                    elif metadata[METADATA_KEY_PROV_TYPE] == PROV_MENTION:
                        # prov mentions used to connect between bundles , see w3c bundle links
                        bundle_records.update({relation_id: DbRelation(attributes, metadata)})

        return list(bundle_records.values())

    def delete_records_by_filter(self, attributes_dict=None, metadata_dict=None):
        """
        Delete a set of records based on filter conditions

        :param attributes_dict: A filter dict with a conjunction of all values in the attributes_dict and metadata_dict
        :type attributes_dict: dict
        :param metadata_dict: A filter for the metadata with a conjunction of all values (also in the attributes_dict )
        :type metadata_dict: dict
        :return: The result of the operation
        :rtype: Bool
        """
        if attributes_dict is None:
            attributes_dict = dict()
        if metadata_dict is None:
            metadata_dict = dict()

        # erase all if no filter set
        if len(attributes_dict) == 0 and len(metadata_dict) == 0:
            del self.all_nodes
            self.all_nodes = dict()
            return True

        # erase only matching nodes
        records_to_delete = self.get_records_by_filter(attributes_dict, metadata_dict)

        for record in records_to_delete:
            if not isinstance(record, DbRecord):
                continue
            identifier = record.metadata[METADATA_KEY_IDENTIFIER]

            if identifier not in self.all_nodes:
                raise NotFoundException("We cant find the id ")
            del self.all_nodes[identifier]

        return True

    def delete_record(self, record_id):
        """
        Delete a single record

        :param record_id: The node id
        :type record_id: str
        :return: Result of the delete operation
        :rtype: Bool
        """

        if record_id not in self.all_nodes:
            raise NotFoundException()

        del self.all_nodes[record_id]

        return True

    def delete_relation(self, relation_id):
        """
        Delete the relation

        :param relation_id: The relation id
        :type relation_id: str
        :return: Result of the delete operation
        :rtype: Bool
        """

        for (from_key, relations) in self.all_relations.items():
            if relation_id in relations:
                del relations[relation_id]
                break

        return True
    @staticmethod
    def _check_attribute_metadata_filter(attributes_filter, metadata_filter, attributes, metadata):
        """
        This function checks if a (attributes, metadata) match to the filter

        :param attributes_dict: A filter dict with a conjunction of all values in the attributes_dict and metadata_dict
        :type attributes_dict: dict
        :param metadata_dict: A filter for the metadata with a conjunction of all values (also in the attributes_dict )
        :type metadata_dict: dict
        :param attributes: The actual provenance data
        :type attributes: dict
        :param metadata: Some metadata that are not PROV-O related
        :type metadata: dict
        :return:
        """
        match = True

        meta_encoded = encode_dict_values_to_primitive(metadata)
        attr_encoded = encode_dict_values_to_primitive(attributes)

        filter_meta_encoded = encode_dict_values_to_primitive(metadata_filter)
        filter_attr_encoded = encode_dict_values_to_primitive(attributes_filter)

        # check properties
        for (key, value) in filter_attr_encoded.items():
            if key not in attr_encoded:
                return False

            elif attr_encoded[key] != value:
                return False

        # check metadata
        for (key, value) in filter_meta_encoded.items():
            if key not in meta_encoded:
                return False

            elif meta_encoded[key] != value:
                return False

        return True



from prov.model import ProvDocument
from provdbconnector import ProvDb
from provdbconnector.db_adapters.in_memory import SimpleInMemoryAdapter

prov_api = ProvDb(adapter=SimpleInMemoryAdapter, auth_info=None)

# create the prov document
prov_document = ProvDocument()
prov_document.add_namespace("ex", "http://example.com")

prov_document.agent("ex:Bob")
prov_document.activity("ex:Alice")

prov_document.influence("ex:Alice", "ex:Bob")

document_id = prov_api.save_document(prov_document)

print(prov_api.get_document_as_provn(document_id))

# Output:
#
# document
#   prefix ex < http: // example.com >
#
#   agent(ex: Bob)
#   activity(ex: Alice, -, -)
#   wasInfluencedBy(ex: Alice, ex: Bob)
# endDocument



from .provapi import ProvDbException


class AdapterException(ProvDbException):
    """
    Base exception class for database adapters.
    """
    pass


class InvalidOptionsException(AdapterException):
    """
    Thrown, if passed argument for adapter is invalid.
    """
    pass


class AuthException(AdapterException):
    """
    Thrown, if database adapter could not establish a connection with given credentials to the database.
    """
    pass


class DatabaseException(AdapterException):
    """
    Thrown, if method could not performed on database.
    """
    pass


class CreateRecordException(DatabaseException):
    """
    Thrown, if record could not be saved in database.
    """
    pass


class CreateRelationException(DatabaseException):
    """
    Thrown, if relation could not be saved in database.
    """
    pass


class NotFoundException(DatabaseException):
    """
    Thrown, if record or relation could not be found in database.
    """
    pass


class MergeException(DatabaseException):
    """
    Thrown, if a record or relation can't get merged
    """
    pass



from prov.tests.examples import primer_example as primer_example, \
    primer_example_alternate, \
    w3c_publication_1, \
    w3c_publication_2, \
    bundles1, \
    bundles2, \
    collections, \
    long_literals, \
    datatypes

from datetime import datetime
from collections import namedtuple

from prov.constants import PROV_RECORD_IDS_MAP, PROV
from prov.model import ProvDocument, ProvActivity, Literal, Identifier
from provdbconnector.db_adapters.baseadapter import METADATA_KEY_NAMESPACES, METADATA_KEY_PROV_TYPE, \
    METADATA_KEY_TYPE_MAP, METADATA_KEY_IDENTIFIER


def prov_db_unknown_prov_typ_example():
    doc = ProvDocument()
    doc.add_namespace("ex", "https://example.com")
    doc.entity(identifier="ex:Entity1")
    doc.entity(identifier="ex:Entity2")
    doc.influence(influencee="ex:Entity1", influencer="ex:Entity2")
    return doc

def prov_default_namespace_example(ns_postfix: str):
    doc = ProvDocument()
    doc.set_default_namespace("https://example.com/{0}".format(ns_postfix))
    doc.entity(identifier="Entity1")
    return doc



def attributes_dict_example():
    """
    Retuns a example dict with some different attributes

    :return: dict with attributes
    :rtype: dict
    """
    attributes = dict()
    attributes.update({"ex:individual attribute": "Some value"})
    attributes.update({"ex:int value": 99})
    attributes.update({"ex:double value": 99.33})
    attributes.update({"ex:date value": datetime.strptime('Jun 1 2005  1:33PM', '%b %d %Y %I:%M%p')})
    attributes.update({"ex:list value": ["list", "of", "strings"]})
    attributes.update({"ex:dict value": {"dict": "value"}})

    return attributes


def base_connector_bundle_parameter_example():
    """
    This example returns a dict with example arguments for a db_adapter

    :return: dict {attributes, metadata}
    :rtype: dict
    """
    doc = ProvDocument()
    doc.add_namespace("ex", "http://example.com")
    attributes = dict()
    attributes.update({"prov:type": "prov:Bundle"})

    namespaces = dict()
    namespaces.update({"ex": "http://example.com"})

    type_map = dict()
    type_map.update({"int value": "int"})
    type_map.update({"date value": "xds:datetime"})

    metadata = dict()

    metadata.update({METADATA_KEY_PROV_TYPE: doc.valid_qualified_name("prov:Entity")})
    metadata.update({METADATA_KEY_IDENTIFIER: doc.valid_qualified_name("ex:bundle name")})
    metadata.update({METADATA_KEY_TYPE_MAP: type_map})
    metadata.update({METADATA_KEY_NAMESPACES: namespaces})

    return_data = dict()
    return_data.update({"attributes": attributes})
    return_data.update({"metadata": metadata})
    return return_data


def base_connector_record_parameter_example():
    """
    Returns a dict with attributes and metadata for a simple node

    :return:dict with attributes metadata
    :rtype: dict
    """
    doc = ProvDocument()

    namespaces = dict()
    namespaces.update({"ex": "http://example.com"})
    namespaces.update({"custom": "http://custom.com"})

    type_map = dict()
    type_map.update({"int value": "int"})
    type_map.update({"date value": "xds:datetime"})

    metadata = dict()

    metadata.update({METADATA_KEY_PROV_TYPE: doc.valid_qualified_name("prov:Activity")})
    metadata.update({METADATA_KEY_IDENTIFIER: doc.valid_qualified_name("prov:example_node")})
    metadata.update({METADATA_KEY_TYPE_MAP: type_map})
    metadata.update({METADATA_KEY_NAMESPACES: namespaces})

    return_data = dict()
    return_data.update({"attributes": attributes_dict_example()})
    return_data.update({"metadata": metadata})

    return return_data


def base_connector_relation_parameter_example():
    """
    Returns a example with a start nodes (attributes, metadata) and also a relation dict with attributes metadata

    :return: dict
    :rtype: dict
    """
    doc = ProvDocument()
    doc.add_namespace("ex", "http://example.com")
    doc.add_namespace("custom", "http://custom.com")

    namespaces = dict()
    namespaces.update({"ex": "http://example.com"})
    namespaces.update({"custom": "http://custom.com"})

    type_map = dict()
    type_map.update({"int value": "int"})
    type_map.update({"date value": "xds:datetime"})

    metadata = dict()

    metadata.update({METADATA_KEY_PROV_TYPE: PROV_RECORD_IDS_MAP["mentionOf"]})
    metadata.update({METADATA_KEY_IDENTIFIER: "identifier for the relation"})
    metadata.update({METADATA_KEY_TYPE_MAP: type_map})
    metadata.update({METADATA_KEY_NAMESPACES: namespaces})

    return_data = dict()
    return_data.update({"attributes": attributes_dict_example()})
    return_data.update({"metadata": metadata})
    return_data.update({"from_node": doc.valid_qualified_name("ex:Yoda")})
    return_data.update({"to_node": doc.valid_qualified_name("ex:Luke Skywalker")})
    return_data.update({"doc": doc})

    return return_data


def base_connector_merge_example():
    """
    This example returns a namedtuple with a from_node relation and to_node
    to test the merge behavior

    :return: namedtuple(from_node, relation, to_node)
    :rtype: namedtuple
    """
    # noinspection PyPep8Naming
    ReturnData = namedtuple("base_connector_merge_example_return_data", "from_node,relation,to_node")
    example_relation = base_connector_relation_parameter_example()

    example_node_a = base_connector_record_parameter_example()
    example_node_b = base_connector_record_parameter_example()

    example_node_a["metadata"][METADATA_KEY_IDENTIFIER] = example_relation["from_node"]
    example_node_b["metadata"][METADATA_KEY_IDENTIFIER] = example_relation["to_node"]

    return ReturnData(example_node_a, example_relation, example_node_b)


def prov_api_record_example():
    """
    This is a more complex record example

    :return:
    """
    doc = ProvDocument()
    doc.add_namespace("ex", "http://example.com")
    doc.add_namespace("custom", "http://custom.com")

    attributes = attributes_dict_example()
    del attributes[
        "ex:dict value"]  # remove dict value because it is not allowed in a prov_record, but for low level adapter tests necessary
    del attributes[
        "ex:list value"]  # remove dict value because it is not allowed in a prov_record, but for low level adapter tests necessary
    attributes.update({"ex:Qualified name ": doc.valid_qualified_name("custom:qualified name")})
    attributes.update({"ex:Qualified name 2": "ex:unqualified_name"})
    attributes.update({"ex:Literal": Literal("test literal", langtag="en")})
    attributes.update({"ex:Literal 2": Literal("test literal with datatype", langtag="en",
                                               datatype=PROV["InternationalizedString"])})
    attributes.update({"ex:identifier type": Identifier("http://example.com/#test")})

    expected_attributes = dict()
    for key, value in attributes.items():
        new_key = doc.valid_qualified_name(key)
        expected_attributes.update({new_key: value})

    # The prov lib don't require to auto convert string values into qualified names
    # valid_name = doc.valid_qualified_name("ex:Qualified name 2")
    # expected_attributes[valid_name] = doc.valid_qualified_name("ex:unqualified_name")

    namespaces = dict()
    namespaces.update({"ex": "http://example.com"})
    namespaces.update({"custom": "http://custom.com"})
    namespaces.update({"prov": "http://www.w3.org/ns/prov#"})

    type_map = dict()
    type_map.update({"ex:date value": {"type": "xsd:dateTime"}})
    type_map.update({"ex:double value": {"type": "xsd:double"}})
    type_map.update({"ex:int value": {"type": "xsd:int"}})

    type_map.update({"ex:Qualified name ": {'type': 'prov:QUALIFIED_NAME'}})
    # type_map.update({"ex:Qualified name 2":{'type': 'prov:QUALIFIED_NAME'}}) #The prov lib don't require to auto convert strings into qualified names
    type_map.update({"ex:Literal": {'lang': 'en'}})
    type_map.update({"ex:Literal 2": {'lang': 'en'}})
    type_map.update({"ex:identifier type": {'type': 'xsd:anyURI'}})

    metadata = dict()
    metadata.update({METADATA_KEY_PROV_TYPE: PROV_RECORD_IDS_MAP["activity"]})
    metadata.update({METADATA_KEY_IDENTIFIER: doc.valid_qualified_name("ex:record")})
    metadata.update({METADATA_KEY_NAMESPACES: namespaces})
    metadata.update({METADATA_KEY_TYPE_MAP: type_map})

    record = ProvActivity(doc, "ex:record", attributes)
    # noinspection PyPep8Naming
    Example = namedtuple("prov_api_metadata_record_example", "metadata, attributes, prov_record, expected_attributes")

    return Example(metadata, attributes, record, expected_attributes)



from prov.model import ProvDocument
from provdbconnector import ProvDb
from provdbconnector import Neo4jAdapter
import os

# create the api

NEO4J_USER = os.environ.get('NEO4J_USERNAME', 'neo4j')
NEO4J_PASS = os.environ.get('NEO4J_PASSWORD', 'neo4jneo4j')
NEO4J_HOST = os.environ.get('NEO4J_HOST', 'localhost')
NEO4J_BOLT_PORT = os.environ.get('NEO4J_BOLT_PORT', '7687')

auth_info = {"user_name": NEO4J_USER,
             "user_password": NEO4J_PASS,
             "host": NEO4J_HOST + ":" + NEO4J_BOLT_PORT
             }

prov_api = ProvDb(adapter=Neo4jAdapter, auth_info=auth_info)

# create the prov document
prov_document = ProvDocument()
prov_document.add_namespace("ex", "http://example.com")

prov_document.agent("ex:Bob")
prov_document.activity("ex:Alice")

prov_document.association("ex:Alice", "ex:Bob")

document_id = prov_api.save_document(prov_document)

print(prov_api.get_document_as_provn(document_id))

# Output:
#
# document
# prefix
# ex < http: // example.com >
#
# agent(ex:Bob)
# activity(ex:Alice, -, -)
# wasAssociatedWith(ex:Alice, ex:Bob, -)
# endDocument
